import re


def parse_roles_text(text: str):
    return [part.strip() for part in re.split(r"[,\n]+", text or "") if part.strip()]


async def start_google_login(request, google_oauth):
    next_url = request.session.get("next") or request.query_params.get("next") or "/"
    request.session["next"] = next_url
    redirect_uri = str(request.base_url).rstrip("/") + "/auth/google/callback"
    return await google_oauth.google.authorize_redirect(request, redirect_uri)


async def fetch_google_userinfo(request, google_oauth, logger):
    token = await google_oauth.google.authorize_access_token(request)
    userinfo = token.get("userinfo")
    if userinfo:
        return userinfo
    try:
        return await google_oauth.google.parse_id_token(request, token)
    except Exception as exc:
        logger.warning(f"Google OAuth id_token missing or invalid: {exc}")
        return await google_oauth.google.userinfo(token=token)


def google_account_allowed(email, google_oauth_cfg):
    allowed_domains = google_oauth_cfg.get("allowed_domains", [])
    if allowed_domains and (not email or email.split("@")[-1] not in allowed_domains):
        return False
    allowed_emails = google_oauth_cfg.get("allowed_emails", [])
    return not allowed_emails or bool(email and email in allowed_emails)


def build_google_auth_payload(userinfo):
    return {
        "provider": "google",
        "email": userinfo.get("email") if isinstance(userinfo, dict) else None,
        "name": userinfo.get("name") if isinstance(userinfo, dict) else None,
        "picture": userinfo.get("picture") if isinstance(userinfo, dict) else None,
    }



from urllib.parse import quote_plus

from fasthtml.common import *
from monsterui.all import *


def render_posts_search_results(query, matches, regex_error):
    trimmed = (query or "").strip()
    if not trimmed:
        return Ul(Li("Type to search file names.", cls="text-[0.7rem] text-center text-slate-500 dark:text-slate-400 bg-transparent"), cls="posts-search-results-list space-y-1 bg-white/0 dark:bg-slate-950/0")
    if not matches:
        return Ul(Li(f'No matches for "{trimmed}".', cls="text-xs text-slate-500 dark:text-slate-400 bg-transparent"), (Li(regex_error, cls="text-[0.7rem] text-center text-amber-600 dark:text-amber-400") if regex_error else None), cls="posts-search-results-list space-y-1 bg-white/0 dark:bg-slate-950/0")
    items, gather_href = [], f"/search/gather?q={quote_plus(trimmed)}"
    items.append(Li(A(Span(UkIcon("layers", cls="w-4 h-4 text-slate-400"), cls="w-4 mr-2 flex items-center justify-center shrink-0"), Span("Gather all search results for LLM", cls="truncate min-w-0 text-xs text-slate-600 dark:text-slate-300"), href=gather_href, hx_get=gather_href, hx_target="#main-content", hx_push_url="true", hx_swap="outerHTML show:window:top settle:0.1s", cls="post-search-link flex items-center py-1 px-2 rounded bg-transparent hover:bg-slate-100 dark:hover:bg-slate-800 text-slate-700 dark:text-slate-300 hover:text-blue-600 transition-colors min-w-0"), cls="bg-transparent"))
    for slug, display in matches:
        items.append(Li(A(Span(UkIcon("search", cls="w-4 h-4 text-slate-400"), cls="w-4 mr-2 flex items-center justify-center shrink-0"), Span(display, cls="truncate min-w-0 font-mono text-xs text-slate-600 dark:text-slate-300", title=display), href=f"/posts/{slug}", hx_get=f"/posts/{slug}", hx_target="#main-content", hx_push_url="true", hx_swap="outerHTML show:window:top settle:0.1s", cls="post-search-link flex items-center py-1 px-2 rounded hover:bg-slate-100 dark:hover:bg-slate-800 text-slate-700 dark:text-slate-300 hover:text-blue-600 transition-colors min-w-0")))
    if regex_error:
        items.append(Li(regex_error, cls="text-[0.7rem] text-center text-amber-600 dark:text-amber-400 mt-1 bg-transparent"))
    return Ul(*items, cls="posts-search-results-list space-y-1 bg-white/0 dark:bg-slate-950/0")


def posts_search_block(initial_results):
    return Div(Div("Filter", cls="text-xs uppercase tracking-widest text-slate-500 dark:text-slate-400 mb-2"), Div(Input(type="search", name="q", placeholder="Search file names…", autocomplete="off", data_placeholder_cycle="1", data_placeholder_primary="Search file names…", data_placeholder_alt="Search regex with /pattern/ syntax", data_search_key="posts", hx_get="/_sidebar/posts/search", hx_trigger="input changed delay:300ms", hx_target="next .posts-search-results", hx_swap="innerHTML", cls="w-full px-3 py-2 text-sm rounded-md border border-slate-200 dark:border-slate-700 bg-white/80 dark:bg-slate-900/60 text-slate-700 dark:text-slate-200 focus:outline-none focus:ring-2 focus:ring-blue-500"), Button("×", type="button", aria_label="Clear search", cls="posts-search-clear-button absolute right-2 top-1/2 -translate-y-1/2 h-6 w-6 rounded text-slate-400 hover:text-slate-600 dark:text-slate-500 dark:hover:text-slate-300 hover:bg-slate-100 dark:hover:bg-slate-800 transition-colors"), cls="relative"), Div(initial_results, id="posts-search-results", cls="posts-search-results mt-4 max-h-64 overflow-y-auto bg-white/0 dark:bg-slate-950/0"), cls="posts-search-block sticky top-0 z-10 bg-white/20 dark:bg-slate-950/70 mb-3")



import hashlib
import re
from pathlib import Path

_CALLOUT_ALIASES = {
    "note": "note", "abstract": "abstract", "summary": "abstract", "tldr": "abstract", "info": "info",
    "todo": "todo", "tip": "tip", "hint": "tip", "important": "tip",
    "success": "success", "check": "success", "done": "success", "question": "question",
    "help": "question", "faq": "question", "warning": "warning", "warn": "warning",
    "caution": "warning", "attention": "warning", "failure": "failure", "fail": "failure",
    "missing": "failure", "danger": "danger", "error": "danger", "bug": "bug",
    "example": "example", "quote": "quote", "cite": "quote", "note": "note",
}


def _placeholder_id(text):
    return hashlib.md5(text.encode()).hexdigest()[:8]


def _strip_one_blockquote_level(text):
    return "\n".join(
        re.sub(r"^\s{0,3}>\s?", "", line) if line.strip() else line
        for line in text.splitlines()
    )


def preprocess_super_sub(content):
    protected = []
    def protect(pattern, text, flags=0):
        regex = re.compile(pattern, flags)
        return regex.sub(lambda m: protected.append(m.group(0)) or f"@@VYASA_PROTECT_{len(protected)-1}@@", text)
    content = protect(r"(```+|~~~+)[\s\S]*?\1", content, re.MULTILINE)
    content = protect(r"(`+)([^`]*?)\1", content)
    content = protect(r"\$\$[\s\S]*?\$\$", content, re.MULTILINE)
    content = protect(r"\$(?:\\.|[^$\n])+\$", content)
    content = protect(r"\\\[[\s\S]*?\\\]", content, re.MULTILINE)
    content = protect(r"\\\((?:\\.|[^\\)])*\\\)", content)
    content = re.sub(r"(?<![\\\w$])\^([A-Za-z0-9.+\-]{1,32})\^(?![\w$])", r"<sup>\1</sup>", content)
    content = re.sub(r"(?<![~\\\w$])~([A-Za-z0-9.+\-]{1,32})~(?![~\w$])", r"<sub>\1</sub>", content)
    for i, chunk in enumerate(protected):
        content = content.replace(f"@@VYASA_PROTECT_{i}@@", chunk)
    return content


def extract_footnotes(content):
    pat = re.compile(r"^\[\^([^\]]+)\]:\s*(.+?)(?=(?:^|\n)\[\^|\n\n|\Z)", re.MULTILINE | re.DOTALL)
    defs = {m.group(1): m.group(2).strip() for m in pat.finditer(content)}
    for m in pat.finditer(content):
        content = content.replace(m.group(0), "", 1)
    return content.strip(), defs


def preserve_newlines(md):
    protected = []
    def protect(pattern, text, flags=0):
        regex = re.compile(pattern, flags)
        return regex.sub(lambda m: protected.append(m.group(0)) or f"__VYASA_BLOCK_{len(protected)-1}__", text)
    md = protect(r"(```+|~~~+)[\s\S]*?\1", md, re.MULTILINE)
    md = protect(r"\$\$[\s\S]*?\$\$", md, re.MULTILINE)
    md = protect(r"\\\[[\s\S]*?\\\]", md, re.MULTILINE)
    md = re.sub(r"(?<!\n)\n(?!\n)", "  \n", md)
    for i, block in enumerate(protected):
        md = md.replace(f"__VYASA_BLOCK_{i}__", block)
    return md


def preprocess_callouts(content):
    callout_store = {}
    protected = []
    content = re.sub(r"(```+|~~~+)[\s\S]*?\1", lambda m: protected.append(m.group(0)) or f"@@VYASA_CALLOUT_BLOCK_{len(protected)-1}@@", content, flags=re.MULTILINE)
    pattern = re.compile(r"^///\s*([a-zA-Z][a-zA-Z0-9_-]*)\s*\n(.*?)^///\s*$", re.MULTILINE | re.DOTALL)
    obsidian_header = re.compile(r"^(\s{0,3}(?:>\s*)+)\[!([A-Za-z][A-Za-z0-9_-]*)\]([+-])?\s*(.*)$")
    quote_line = re.compile(r"^(\s{0,3}(?:>\s*)+)(.*)$")

    def replace(match):
        kind = _CALLOUT_ALIASES.get(match.group(1).strip().lower(), match.group(1).strip().lower())
        body = match.group(2).strip()
        if not body:
            return match.group(0)
        callout_id = _placeholder_id(match.group(0))
        callout_store[callout_id] = {"kind": kind, "body": body, "title": None, "fold": None}
        return f'\n\n<div class="vyasa-callout-placeholder" data-callout-id="{callout_id}"></div>\n\n'

    content = pattern.sub(replace, content)
    lines, out, i = content.splitlines(), [], 0
    while i < len(lines):
        header = obsidian_header.match(lines[i])
        if not header:
            out.append(lines[i]); i += 1; continue
        depth = header.group(1).count(">")
        body, i = [], i + 1
        while i < len(lines):
            if not lines[i].strip():
                body.append("")
                i += 1
                continue
            quoted = quote_line.match(lines[i])
            if not quoted:
                break
            line_depth = quoted.group(1).count(">")
            if line_depth < depth:
                break
            if line_depth == depth and obsidian_header.match(lines[i]):
                break
            body.append(lines[i].split(">", 1)[1].lstrip() if line_depth == depth else quoted.group(0).split(">", 1)[1].lstrip())
            i += 1
        callout_id = _placeholder_id("\n".join([lines[i - len(body) - 1], *body]))
        callout_store[callout_id] = {"kind": _CALLOUT_ALIASES.get(header.group(2).lower(), header.group(2).lower()), "body": "\n".join(body).strip(), "title": header.group(4).strip() or None, "fold": header.group(3) or None}
        out.extend(["", f'<div class="vyasa-callout-placeholder" data-callout-id="{callout_id}"></div>', ""])
    content = "\n".join(out)
    for item in callout_store.values():
        for i, block in enumerate(protected):
            item["body"] = item["body"].replace(
                f"@@VYASA_CALLOUT_BLOCK_{i}@@", _strip_one_blockquote_level(block)
            )
    for i, block in enumerate(protected):
        content = content.replace(f"@@VYASA_CALLOUT_BLOCK_{i}@@", block)
    return content, callout_store


def preprocess_code_includes(content, current_path=None, root_folder=None):
    protected = []
    def protect(match):
        protected.append(match.group(0))
        return f"@@VYASA_CODE_BLOCK_{len(protected)-1}@@"

    include_store = {}
    pattern = re.compile(r"^\{\*\s+(.+?)\s+\*\}\s*$", re.MULTILINE)
    base_dir = (Path(root_folder) / Path(current_path).parent) if current_path and root_folder else None
    content = re.sub(r"(```+|~~~+)[\s\S]*?\1", protect, content, flags=re.MULTILINE)

    def replace(match):
        spec = match.group(1).strip()
        path_text = spec.split(" ln[", 1)[0].split(" hl[", 1)[0].strip()
        file_path = (base_dir / path_text).resolve() if base_dir else Path(path_text).resolve()
        include_id = _placeholder_id(match.group(0))
        include_store[include_id] = {"spec": spec, "path_text": path_text, "file_path": file_path}
        return f'<div class="vyasa-code-include-placeholder" data-include-id="{include_id}"></div>'

    content = pattern.sub(replace, content)
    for i, block in enumerate(protected):
        content = content.replace(f"@@VYASA_CODE_BLOCK_{i}@@", block)
    return content, include_store



import json, os, re
from pathlib import Path


def normalize_rbac_cfg(cfg, coerce_list):
    cfg = cfg if isinstance(cfg, dict) else {}
    user_roles = cfg.get("user_roles", {})
    role_users = cfg.get("role_users", {})
    rules = cfg.get("rules", [])
    cleaned_rules = []
    for rule in rules if isinstance(rules, list) else []:
        if not isinstance(rule, dict):
            continue
        pattern = rule.get("pattern")
        roles = coerce_list(rule.get("roles", []))
        if pattern and roles:
            cleaned_rules.append({"pattern": str(pattern), "roles": roles})
    return {
        "enabled": bool(cfg.get("enabled", False)),
        "default_roles": coerce_list(cfg.get("default_roles", [])),
        "user_roles": user_roles if isinstance(user_roles, dict) else {},
        "role_users": role_users if isinstance(role_users, dict) else {},
        "rules": cleaned_rules,
    }


def render_rbac_toml(cfg, coerce_list):
    cfg = normalize_rbac_cfg(cfg, coerce_list)
    q = lambda value: json.dumps(str(value))
    lst = lambda items: "[" + ", ".join(q(item) for item in items) + "]"
    table = lambda mapping: "{}" if not mapping else "{ " + ", ".join(
        f"{q(key)} = {lst(coerce_list(mapping[key]))}" for key in sorted(mapping.keys())
    ) + " }"
    lines = ["[rbac]", f"enabled = {'true' if cfg.get('enabled') else 'false'}", f"default_roles = {lst(cfg.get('default_roles', []))}", f"user_roles = {table(cfg.get('user_roles', {}))}", f"role_users = {table(cfg.get('role_users', {}))}", ""]
    for rule in cfg.get("rules", []):
        lines.extend(["[[rbac.rules]]", f"pattern = {q(rule.get('pattern'))}", f"roles = {lst(rule.get('roles', []))}", ""])
    return "\n".join(lines).rstrip() + "\n"


def write_rbac_to_vyasa(cfg, coerce_list, root_folder):
    root_env = os.getenv("VYASA_ROOT")
    path = Path(root_env) / ".vyasa" if root_env and (Path(root_env) / ".vyasa").exists() else (Path.cwd() / ".vyasa" if (Path.cwd() / ".vyasa").exists() else root_folder / ".vyasa")
    text = path.read_text(encoding="utf-8") if path.exists() else ""
    new_block = render_rbac_toml(cfg, coerce_list)
    pattern = r"(?ms)^\[rbac\]\n.*?(?=^\[[^\[]|\Z)"
    text = re.sub(pattern, new_block + "\n", text) if re.search(r"(?m)^\[rbac\]", text) else ((text + ("" if not text or text.endswith("\n") else "\n")) + "\n" + new_block)
    path.write_text(text, encoding="utf-8")



from fasthtml.common import *
from monsterui.all import *


def not_found_content():
    return Div(
        Div(H1("404", cls="text-9xl font-bold text-slate-300 dark:text-slate-700 mb-4"), cls="text-center"),
        H2("Page Not Found", cls="text-3xl font-bold text-slate-800 dark:text-slate-200 mb-4 text-center"),
        P("Oops! The page you're looking for doesn't exist. It might have been moved or deleted.", cls="text-lg text-slate-600 dark:text-slate-400 mb-8 text-center max-w-2xl mx-auto"),
        Div(
            A(UkIcon("home", cls="w-5 h-5 mr-2"), "Go to Home", href="/", hx_get="/", hx_target="#main-content", hx_push_url="true", hx_swap="outerHTML show:window:top settle:0.1s", cls="inline-flex items-center px-6 py-3 bg-blue-600 hover:bg-blue-700 text-white rounded-lg font-medium transition-colors mr-4"),
            A(UkIcon("arrow-left", cls="w-5 h-5 mr-2"), "Go Back", href="javascript:history.back()", cls="inline-flex items-center px-6 py-3 bg-slate-200 hover:bg-slate-300 dark:bg-slate-700 dark:hover:bg-slate-600 text-slate-800 dark:text-slate-200 rounded-lg font-medium transition-colors"),
            cls="flex justify-center items-center gap-4 flex-wrap",
        ),
        Div(P("💡 ", Strong("Tip:"), " Check the sidebar for available posts, or use the search to find what you're looking for.", cls="text-sm text-slate-500 dark:text-slate-500 italic"), cls="mt-12 text-center"),
        cls="flex flex-col items-center justify-center py-16 px-6 min-h-[60vh]",
    )



from pathlib import Path

from .helpers import (
    _effective_ignore_list,
    _effective_include_list,
    _should_include_folder,
    find_folder_note_file,
    get_vyasa_config,
    order_vyasa_entries,
    should_exclude_dir,
)


def get_tree_entries(folder: Path, root: Path, show_hidden: bool, excluded_dirs: set[str], allowed_suffixes: tuple[str, ...]) -> list[Path]:
    index_file = None
    if folder == root:
        for stem in ("index", "readme"):
            for candidate in root.iterdir():
                if candidate.is_file() and candidate.suffix == ".md" and candidate.stem.lower() == stem:
                    index_file = candidate
                    break
            if index_file:
                break
    folder_note = find_folder_note_file(folder)
    ignore_list = _effective_ignore_list(root, folder)
    include_list = _effective_include_list(root, folder)
    entries = []
    for item in folder.iterdir():
        if item.name == ".vyasa":
            continue
        if item.is_dir():
            if should_exclude_dir(item.name, excluded_dirs) or (not show_hidden and item.name.startswith(".")):
                continue
            if _should_include_folder(item.name, include_list, ignore_list):
                entries.append(item)
        elif item.suffix in allowed_suffixes:
            if (folder_note and item.resolve() == folder_note.resolve()) or (index_file and item.resolve() == index_file.resolve()):
                continue
            entries.append(item)
    return order_vyasa_entries(entries, get_vyasa_config(folder))



from fasthtml.common import Div, H1, P


def gather_search_page(htmx, q="", request=None, *, find_search_matches, get_roles_from_request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list, get_root_folder, is_allowed, gather_search_content, layout):
    matches, regex_error = find_search_matches(q, limit=200)
    roles = get_roles_from_request(request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list)
    if roles is not None:
        root = get_root_folder()
        matches = [item for item in matches if is_allowed(f"/posts/{item.relative_to(root).with_suffix('')}", roles or [], rbac_rules)]
    if not matches:
        content = Div(H1("Search Results", cls="text-3xl font-bold mb-6"), P("No matching posts found.", cls="text-slate-600 dark:text-slate-400"), P(regex_error, cls="text-amber-600 dark:text-amber-400 text-sm") if regex_error else None)
        return layout(content, htmx=htmx, title="Search Results", show_sidebar=True, auth=request.scope.get("auth") if request else None)
    return layout(gather_search_content(q, matches, regex_error, get_root_folder()), htmx=htmx, title="Search Results", show_sidebar=True, auth=request.scope.get("auth") if request else None)



from .policy import normalize_auth, resolve_roles


def get_auth_from_request(request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list):
    if not request:
        return None
    auth = None
    try:
        auth = request.scope.get("auth")
    except Exception:
        auth = None
    if not auth:
        try:
            auth = request.session.get("auth")
        except Exception:
            auth = None
    auth = normalize_auth(auth) if auth else None
    if auth and rbac_rules:
        auth["roles"] = auth.get("roles") or resolve_roles(auth, rbac_cfg, google_oauth_cfg, coerce_list)
    return auth


def get_roles_from_request(request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list):
    auth = get_auth_from_request(request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list)
    return auth.get("roles") if auth else []


def get_roles_from_auth(auth, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list):
    auth = normalize_auth(auth) if auth else None
    if auth and rbac_rules:
        auth["roles"] = auth.get("roles") or resolve_roles(auth, rbac_cfg, google_oauth_cfg, coerce_list)
    return auth.get("roles") if auth else []



import time
from urllib.parse import quote

from fasthtml.common import A, Aside, Button, Div, Footer, Main, NotStr, P, Span, Title
from monsterui.all import UkIcon


def render_layout(*content, htmx, title=None, show_sidebar=False, toc_content=None, current_path=None, show_toc=True, auth=None, htmx_nav=True, nav_posts_menu=False, full_width=False, show_footer=True, no_scroll=False, logger, resolve_layout_config, width_class_and_style, style_attr, get_sidebar_custom_css_links, get_root_folder, build_sidebar_toc_items, extract_sidebar_toc, strip_inline_markdown, text_to_anchor, unique_anchor, get_config, build_collapsible_sidebar, get_roles_from_auth, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list, cached_posts_sidebar_html, posts_sidebar_fingerprint, get_posts, navbar):
    layout_start_time = time.time()
    logger.debug("[LAYOUT] layout() start")
    section_class = f"section-{current_path.replace('/', '-')}" if current_path else ""
    t_section = time.time()
    logger.debug(f"[LAYOUT] section_class computed in {(t_section - layout_start_time)*1000:.2f}ms")
    layout_config = resolve_layout_config(current_path)
    layout_max_class, layout_max_style = width_class_and_style(layout_config.get("layout_max_width"), "max")
    layout_fluid_class = "layout-fluid" if layout_max_style else ""
    if full_width:
        layout_max_class = layout_max_style = layout_fluid_class = ""
    main_spacing_cls = "px-0 py-0" if no_scroll else "px-6 py-8"
    page_container_cls = "flex flex-col h-screen overflow-hidden" if no_scroll else "flex flex-col min-h-screen"
    navbar_margin_cls = "mt-0" if no_scroll else "mt-4"

    def _footer_node(outer_cls, outer_style):
        logout_button = None
        if auth:
            display_name = auth.get("name") or auth.get("email") or auth.get("username") or "User"
            impersonator = auth.get("impersonator")
            if impersonator:
                original = impersonator.get("name") or impersonator.get("email") or impersonator.get("username") or "User"
                display_name = f"Impersonating {display_name} (as {original})"
            logout_button = A(f"Logout {display_name}", href="/logout", cls="text-sm text-white/80 hover:text-white underline")
        footer_inner = Div(Div(logout_button, cls="flex items-center") if logout_button else Div(), Div(NotStr('Powered by <a href="https://github.com/sizhky/vyasa" class="underline hover:text-white/80" target="_blank" rel="noopener noreferrer">Vyasa</a> and ❤️')), cls="flex items-center justify-between w-full")
        return Footer(Div(footer_inner, cls="bg-slate-900 text-white rounded-lg p-4 my-4 dark:bg-slate-800"), cls=outer_cls, id="site-footer", **outer_style)

    if htmx and getattr(htmx, "request", None):
        return _render_htmx_layout(content, title, show_sidebar, toc_content, current_path, show_toc, logger, t_section, layout_start_time, main_spacing_cls, section_class, get_sidebar_custom_css_links, get_root_folder, build_sidebar_toc_items, extract_sidebar_toc, strip_inline_markdown, text_to_anchor, unique_anchor, get_config, build_collapsible_sidebar)

    return _render_full_layout(content, title, show_sidebar, toc_content, current_path, show_toc, auth, htmx_nav, nav_posts_menu, show_footer, no_scroll, logger, t_section, layout_start_time, layout_fluid_class, layout_max_class, layout_max_style, main_spacing_cls, page_container_cls, navbar_margin_cls, section_class, get_sidebar_custom_css_links, get_root_folder, build_sidebar_toc_items, extract_sidebar_toc, strip_inline_markdown, text_to_anchor, unique_anchor, get_config, build_collapsible_sidebar, get_roles_from_auth, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list, cached_posts_sidebar_html, posts_sidebar_fingerprint, get_posts, navbar, style_attr, _footer_node)


def _toc_items(toc_content, build_sidebar_toc_items, extract_sidebar_toc, strip_inline_markdown, text_to_anchor, unique_anchor):
    return build_sidebar_toc_items(extract_sidebar_toc(toc_content, strip_inline_markdown, text_to_anchor, unique_anchor)) if toc_content else []


def _render_htmx_layout(content, title, show_sidebar, toc_content, current_path, show_toc, logger, t_section, layout_start_time, main_spacing_cls, section_class, get_sidebar_custom_css_links, get_root_folder, build_sidebar_toc_items, extract_sidebar_toc, strip_inline_markdown, text_to_anchor, unique_anchor, get_config, build_collapsible_sidebar):
    if show_sidebar:
        toc_sidebar = None
        t_toc = t_section
        if show_toc:
            toc_items = _toc_items(toc_content, build_sidebar_toc_items, extract_sidebar_toc, strip_inline_markdown, text_to_anchor, unique_anchor)
            t_toc = time.time()
            logger.debug(f"[LAYOUT] TOC built in {(t_toc - t_section)*1000:.2f}ms")
            sidebars_open = get_config().get_sidebars_open()
            toc_sidebar = Aside(build_collapsible_sidebar("list", "Table of Contents", toc_items, is_open=sidebars_open, shortcut_key="X") if toc_items else Div(), cls="hidden xl:block w-72 shrink-0 sticky top-24 self-start max-h-[calc(100vh-10rem)] overflow-hidden z-[1000]", id="toc-sidebar", hx_swap_oob="true")
            mobile_toc_panel = Div(Div(Button(UkIcon("x", cls="w-5 h-5"), id="close-mobile-toc", cls="p-2 hover:bg-slate-200 dark:hover:bg-slate-700 rounded transition-colors ml-auto", type="button"), cls="flex justify-end p-2 bg-white dark:bg-slate-950 border-b border-slate-200 dark:border-slate-800"), Div(build_collapsible_sidebar("list", "Table of Contents", toc_items, is_open=sidebars_open, shortcut_key="X") if toc_items else Div(P("No table of contents available.", cls="text-slate-500 dark:text-slate-400 text-sm p-4")), cls="p-4 overflow-y-auto"), id="mobile-toc-panel", cls="fixed inset-0 bg-white dark:bg-slate-950 z-[9999] xl:hidden transform translate-x-full transition-transform duration-300", hx_swap_oob="true")
        custom_css_links = get_sidebar_custom_css_links(get_root_folder(), current_path, section_class)
        logger.debug(f"[LAYOUT] Custom CSS resolved in {(time.time() - t_toc)*1000:.2f}ms")
        main_content_container = Main(*content, cls=f"flex-1 min-w-0 {main_spacing_cls} space-y-8 {section_class}", id="main-content", hx_boost="true", hx_target="#main-content", hx_swap="outerHTML show:window:top settle:0.1s")
        result = [Title(title), Div(*custom_css_links, id="scoped-css-container", hx_swap_oob="true") if custom_css_links else Div(id="scoped-css-container", hx_swap_oob="true")]
        if show_toc:
            result.append(mobile_toc_panel)
        if toc_sidebar:
            result.extend([main_content_container, toc_sidebar])
        else:
            result.extend([main_content_container, Div(id="toc-sidebar", hx_swap_oob="true"), Div(id="mobile-toc-panel", hx_swap_oob="true")])
        logger.debug(f"[LAYOUT] TOTAL layout() time {(time.time() - layout_start_time)*1000:.2f}ms")
        return tuple(result)
    custom_css_links = get_sidebar_custom_css_links(get_root_folder(), current_path, section_class) if current_path else []
    result = [Title(title), Div(*custom_css_links, id="scoped-css-container", hx_swap_oob="true") if custom_css_links else Div(id="scoped-css-container", hx_swap_oob="true"), *content]
    logger.debug(f"[LAYOUT] TOTAL layout() time {(time.time() - layout_start_time)*1000:.2f}ms")
    return tuple(result)


def _render_full_layout(content, title, show_sidebar, toc_content, current_path, show_toc, auth, htmx_nav, nav_posts_menu, show_footer, no_scroll, logger, t_section, layout_start_time, layout_fluid_class, layout_max_class, layout_max_style, main_spacing_cls, page_container_cls, navbar_margin_cls, section_class, get_sidebar_custom_css_links, get_root_folder, build_sidebar_toc_items, extract_sidebar_toc, strip_inline_markdown, text_to_anchor, unique_anchor, get_config, build_collapsible_sidebar, get_roles_from_auth, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list, cached_posts_sidebar_html, posts_sidebar_fingerprint, get_posts, navbar, style_attr, footer_node):
    if show_sidebar:
        toc_sidebar = None
        if show_toc:
            toc_items = _toc_items(toc_content, build_sidebar_toc_items, extract_sidebar_toc, strip_inline_markdown, text_to_anchor, unique_anchor)
            logger.debug(f"[LAYOUT] TOC built in {(time.time() - t_section)*1000:.2f}ms")
            sidebars_open = get_config().get_sidebars_open()
            toc_sidebar = Aside(build_collapsible_sidebar("list", "Table of Contents", toc_items, is_open=sidebars_open, shortcut_key="X") if toc_items else Div(), cls="hidden xl:block w-72 shrink-0 sticky top-24 self-start max-h-[calc(100vh-10rem)] overflow-hidden z-[1000]", id="toc-sidebar")
        custom_css_links = get_sidebar_custom_css_links(get_root_folder(), current_path, section_class)
        main_content_container = Main(*content, cls=f"flex-1 min-w-0 {main_spacing_cls} space-y-8 {section_class}", id="main-content", hx_boost="true", hx_target="#main-content", hx_swap="outerHTML show:window:top settle:0.1s")
        roles_key = tuple(get_roles_from_auth(auth, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list) or [])
        mobile_posts_panel = Div(Div(Button(UkIcon("x", cls="w-5 h-5"), id="close-mobile-posts", cls="p-2 hover:bg-slate-200 dark:hover:bg-slate-700 rounded transition-colors ml-auto", type="button"), cls="flex justify-end p-2 bg-white dark:bg-slate-950 border-b border-slate-200 dark:border-slate-800"), Div(NotStr(cached_posts_sidebar_html(posts_sidebar_fingerprint(), roles_key, get_config().get_show_hidden(), current_path or "")), cls="p-4 overflow-y-auto"), id="mobile-posts-panel", cls="fixed inset-0 bg-white dark:bg-slate-950 z-[9999] xl:hidden transform -translate-x-full transition-transform duration-300")
        mobile_toc_panel = Div(Div(Button(UkIcon("x", cls="w-5 h-5"), id="close-mobile-toc", cls="p-2 hover:bg-slate-200 dark:hover:bg-slate-700 rounded transition-colors ml-auto", type="button"), cls="flex justify-end p-2 bg-white dark:bg-slate-950 border-b border-slate-200 dark:border-slate-800"), Div(build_collapsible_sidebar("list", "Table of Contents", toc_items, is_open=sidebars_open, shortcut_key="X") if (show_toc and toc_items) else Div(P("No table of contents available.", cls="text-slate-500 dark:text-slate-400 text-sm p-4")), cls="p-4 overflow-y-auto"), id="mobile-toc-panel", cls="fixed inset-0 bg-white dark:bg-slate-950 z-[9999] xl:hidden transform translate-x-full transition-transform duration-300") if show_toc else None
        nav_posts_items = get_posts(list(roles_key) if roles_key else [], current_path=current_path or "") if nav_posts_menu else None
        content_with_sidebars = Div(cls=f"layout-container {layout_fluid_class} w-full {layout_max_class} mx-auto px-4 flex gap-6 flex-1 {'min-h-0' if no_scroll else ''}".strip(), id="content-with-sidebars", **style_attr(layout_max_style))((Aside(Div(UkIcon("loader", cls="w-5 h-5 animate-spin"), Span("Loading posts…", cls="ml-2 text-sm"), cls="flex items-center justify-center h-32 text-slate-400"), cls="hidden xl:block w-72 shrink-0 sticky top-24 self-start max-h-[calc(100vh-10rem)] overflow-x-auto overflow-y-hidden z-[1000]", id="posts-sidebar", hx_get=f"/_sidebar/posts?current_path={quote(current_path or '', safe='')}", hx_trigger="load", hx_swap="outerHTML") if not nav_posts_menu else None), main_content_container, toc_sidebar if toc_sidebar else None)
        body_content = Div(id="page-container", cls=page_container_cls)(Div(navbar(show_mobile_menus=True, htmx_nav=htmx_nav, posts_menu_items=nav_posts_items, compact_mode=nav_posts_menu), cls=f"layout-container {layout_fluid_class} w-full {layout_max_class} mx-auto px-4 sticky top-0 z-50 {navbar_margin_cls}".strip(), id="site-navbar", **style_attr(layout_max_style)), mobile_posts_panel, mobile_toc_panel if mobile_toc_panel else None, content_with_sidebars, footer_node(f"layout-container {layout_fluid_class} w-full {layout_max_class} mx-auto px-6 mt-auto mb-6".strip(), style_attr(layout_max_style)) if show_footer else None)
    else:
        custom_css_links = get_sidebar_custom_css_links(get_root_folder(), current_path, section_class) if current_path else []
        body_content = Div(id="page-container", cls="flex flex-col min-h-screen")(Div(navbar(htmx_nav=htmx_nav), cls=f"layout-container {layout_fluid_class} w-full {layout_max_class} mx-auto px-4 sticky top-0 z-50 mt-4".strip(), id="site-navbar", **style_attr(layout_max_style)), Main(*content, cls=f"layout-container {layout_fluid_class} w-full {layout_max_class} mx-auto px-6 py-8 space-y-8".strip(), id="main-content", hx_boost="true", hx_target="#main-content", hx_swap="outerHTML show:window:top settle:0.1s", **style_attr(layout_max_style)), footer_node(f"layout-container {layout_fluid_class} w-full {layout_max_class} mx-auto px-6 mt-auto mb-6".strip(), style_attr(layout_max_style)) if show_footer else None)
    result = [Title(title), Div(*custom_css_links, id="scoped-css-container") if custom_css_links else Div(id="scoped-css-container"), body_content]
    logger.debug(f"[LAYOUT] FULL PAGE assembled in {(time.time() - layout_start_time)*1000:.2f}ms")
    return tuple(result)



def build_google_oauth(google_oauth_cfg, logger):
    oauth = None
    enabled = False
    if google_oauth_cfg.get("client_id") and google_oauth_cfg.get("client_secret"):
        try:
            from authlib.integrations.starlette_client import OAuth

            oauth = OAuth()
            oauth.register(
                name="google",
                client_id=google_oauth_cfg["client_id"],
                client_secret=google_oauth_cfg["client_secret"],
                server_metadata_url="https://accounts.google.com/.well-known/openid-configuration",
                userinfo_endpoint="https://openidconnect.googleapis.com/v1/userinfo",
                client_kwargs={"scope": "openid email profile"},
            )
            enabled = True
        except Exception as exc:
            logger.warning(f"Google OAuth disabled: {exc}")
    return oauth, enabled



import os
import tomllib


def drawing_password_for(root, path: str):
    if not path:
        return None
    drawing = (root.resolve() / f"{path.strip('/')}.excalidraw").resolve()
    if not str(drawing).startswith(str(root.resolve()) + os.sep):
        return None
    rel_root = drawing.with_suffix("").relative_to(root).as_posix()
    cur = drawing.parent
    while True:
        cfg = cur / ".vyasa"
        if cfg.exists():
            try:
                passwords = tomllib.loads(cfg.read_text(encoding="utf-8")).get("drawings_passwords", {})
                if isinstance(passwords, dict):
                    rel_cur = drawing.with_suffix("").relative_to(cur).as_posix()
                    if rel_cur in passwords: return str(passwords[rel_cur])
                    if rel_root in passwords: return str(passwords[rel_root])
            except Exception:
                pass
        if cur == root or cur.parent == cur:
            break
        cur = cur.parent
    return None


def drawing_unlocked_in_session(session, path: str) -> bool:
    return bool((session or {}).get("drawings_unlocked", {}).get(path.strip("/")))


def unlock_drawing(session, path: str):
    key = path.strip("/")
    unlocked = dict(session.get("drawings_unlocked", {}))
    unlocked[key] = True
    session["drawings_unlocked"] = unlocked



import json
from dataclasses import dataclass

from fastsql import Database


@dataclass
class RbacConfigRow:
    key: str
    value: str


def get_rbac_table(root_folder, cache, create_if_missing=True):
    if cache["db"] is None:
        db_path = root_folder / ".vyasa-rbac.db"
        if not create_if_missing and not db_path.exists():
            return None, None
        db_path.parent.mkdir(parents=True, exist_ok=True)
        cache["db"] = Database(f"sqlite:///{db_path}")
        cache["tbl"] = cache["db"].create(RbacConfigRow, pk="key", name="rbac_config")
    return cache["db"], cache["tbl"]


def load_rbac_cfg(root_folder, cache, normalize):
    _, tbl = get_rbac_table(root_folder, cache, create_if_missing=False)
    if tbl is None:
        return None
    rows = tbl()
    if not rows:
        return None
    data = {}
    for row in rows:
        data[row.key] = json.loads(row.value)
    return normalize(data)


def write_rbac_cfg(root_folder, cache, cfg, normalize):
    _, tbl = get_rbac_table(root_folder, cache)
    cfg = normalize(cfg)
    existing = {row.key for row in tbl()}
    for key, value in cfg.items():
        payload = json.dumps(value, sort_keys=True)
        tbl.update(key=key, value=payload) if key in existing else tbl.insert(RbacConfigRow(key=key, value=payload))
    for key in existing - set(cfg.keys()):
        tbl.delete(key)



def normalize_auth(auth):
    if not auth:
        return None
    if isinstance(auth, dict):
        return auth
    return {"provider": "local", "username": str(auth)}


def path_requires_roles(path, rbac_rules):
    for pattern, _roles in rbac_rules:
        if pattern.search(path):
            return True
    return False


def is_allowed(path, roles, rbac_rules):
    if not rbac_rules:
        return True
    roles_set = set(roles or [])
    matched_any = False
    allowed = False
    for pattern, allowed_roles in rbac_rules:
        if pattern.search(path):
            matched_any = True
            if roles_set & allowed_roles:
                allowed = True
    return allowed if matched_any else True


def resolve_roles(auth, rbac_cfg, google_oauth_cfg, coerce_list):
    auth = normalize_auth(auth) or {}
    username = auth.get("username")
    email = auth.get("email")
    user_roles = rbac_cfg.get("user_roles", {})
    roles = []
    if isinstance(user_roles, dict):
        if email and email in user_roles:
            roles.extend(coerce_list(user_roles.get(email)))
        if username and username in user_roles:
            roles.extend(coerce_list(user_roles.get(username)))
    role_users = rbac_cfg.get("role_users", {})
    if isinstance(role_users, dict):
        for role, users in role_users.items():
            users_list = coerce_list(users)
            if email and email in users_list:
                roles.append(role)
            if username and username in users_list:
                roles.append(role)
    if not roles:
        roles = rbac_cfg.get("default_roles", []) or google_oauth_cfg.get("default_roles", [])
    roles = [r for r in roles if r]
    return list(dict.fromkeys(roles)) if roles else []



import time
from urllib.parse import quote

from fasthtml.common import A, Details, Li, Span, Summary, Ul
from monsterui.all import UkIcon


def _folder_summary(title_node, branch_href=None):
    kwargs = {}
    if branch_href:
        kwargs = {"hx_get": branch_href, "hx_trigger": "click once", "hx_target": "next ul", "hx_swap": "innerHTML"}
    return Summary(
        Span(Span(cls="folder-chevron"), cls="w-4 mr-2 flex items-center justify-center shrink-0"),
        Span(UkIcon("folder", cls="text-blue-500 w-4 h-4"), cls="w-4 mr-2 flex items-center justify-center shrink-0"),
        title_node,
        cls="inline-flex w-max items-center font-medium cursor-pointer py-1 px-2 hover:text-blue-600 select-none list-none rounded hover:bg-slate-100 dark:hover:bg-slate-800 transition-colors whitespace-nowrap",
        **kwargs,
    )


def folder_has_visible_descendant(folder, roles, depth, *, root, show_hidden, excluded_dirs, get_nav_entries, is_allowed_fn, rbac_rules):
    for item in get_nav_entries(folder, root, show_hidden, excluded_dirs):
        if item.is_dir():
            if depth > 0 and folder_has_visible_descendant(item, roles, depth - 1, root=root, show_hidden=show_hidden, excluded_dirs=excluded_dirs, get_nav_entries=get_nav_entries, is_allowed_fn=is_allowed_fn, rbac_rules=rbac_rules):
                return True
            continue
        slug = str(item.relative_to(root).with_suffix(""))
        route = f"/drawings/{slug}" if item.suffix == ".excalidraw" else f"/posts/{slug}"
        if is_allowed_fn(route, roles or [], rbac_rules):
            return True
    return False


def build_post_tree_render(folder, roles=None, max_depth=None, active_parts=(), *, root, show_hidden, excluded_dirs, get_nav_entries, effective_abbreviations, should_exclude_dir_fn, slug_to_title_fn, find_folder_note_file_fn, is_allowed_fn, parse_frontmatter_fn, rbac_rules, logger):
    items, start_time = [], time.time()
    try:
        entries = get_nav_entries(folder, root, show_hidden, excluded_dirs)
        abbreviations = effective_abbreviations(root, folder)
    except (OSError, PermissionError):
        return items
    for item in entries:
        if item.is_dir():
            if should_exclude_dir_fn(item.name, excluded_dirs) or (not show_hidden and item.name.startswith(".")):
                continue
            folder_title = slug_to_title_fn(item.name, abbreviations=abbreviations)
            rel_folder = str(item.relative_to(root))
            child_active = tuple(active_parts[1:]) if active_parts and active_parts[0] == item.name else ()
            should_expand = bool(active_parts and active_parts[0] == item.name)
            if max_depth == 0:
                if should_expand:
                    sub_items = build_post_tree_render(item, roles=roles, max_depth=0 if not child_active else None, active_parts=child_active, root=root, show_hidden=show_hidden, excluded_dirs=excluded_dirs, get_nav_entries=get_nav_entries, effective_abbreviations=effective_abbreviations, should_exclude_dir_fn=should_exclude_dir_fn, slug_to_title_fn=slug_to_title_fn, find_folder_note_file_fn=find_folder_note_file_fn, is_allowed_fn=is_allowed_fn, parse_frontmatter_fn=parse_frontmatter_fn, rbac_rules=rbac_rules, logger=logger)
                    items.append(Li(Details(_folder_summary(Span(folder_title, cls="whitespace-nowrap", title=folder_title)), Ul(*sub_items, cls="ml-4 pl-2 space-y-1 border-l border-slate-100 dark:border-slate-800"), data_folder="true", open=True), cls="my-1"))
                    continue
                if not folder_has_visible_descendant(item, roles, 3, root=root, show_hidden=show_hidden, excluded_dirs=excluded_dirs, get_nav_entries=get_nav_entries, is_allowed_fn=is_allowed_fn, rbac_rules=rbac_rules):
                    continue
                branch_href = f"/_sidebar/posts/branch?path={quote(rel_folder, safe='')}"
                items.append(Li(Details(_folder_summary(Span(folder_title, cls="whitespace-nowrap", title=folder_title), branch_href=branch_href), Ul(cls="ml-4 pl-2 space-y-1 border-l border-slate-100 dark:border-slate-800"), data_folder="true"), cls="my-1"))
                continue
            sub_items = build_post_tree_render(item, roles=roles, max_depth=None if should_expand else (None if max_depth is None else max_depth - 1), active_parts=child_active, root=root, show_hidden=show_hidden, excluded_dirs=excluded_dirs, get_nav_entries=get_nav_entries, effective_abbreviations=effective_abbreviations, should_exclude_dir_fn=should_exclude_dir_fn, slug_to_title_fn=slug_to_title_fn, find_folder_note_file_fn=find_folder_note_file_fn, is_allowed_fn=is_allowed_fn, parse_frontmatter_fn=parse_frontmatter_fn, rbac_rules=rbac_rules, logger=logger)
            note_file = find_folder_note_file_fn(item)
            note_slug = str(note_file.relative_to(root).with_suffix("")) if note_file else None
            note_allowed = bool(note_slug and is_allowed_fn(f"/posts/{note_slug}", roles or [], rbac_rules))
            note_link = A(href=f"/posts/{note_slug}", hx_get=f"/posts/{note_slug}", hx_target="#main-content", hx_push_url="true", hx_swap="outerHTML show:window:top settle:0.1s", cls="folder-note-link whitespace-nowrap hover:underline", title=f"Open {folder_title}", onclick="event.stopPropagation();")(folder_title) if note_allowed else None
            if not sub_items and not note_allowed:
                continue
            if sub_items:
                items.append(Li(Details(_folder_summary(note_link or Span(folder_title, cls="whitespace-nowrap", title=folder_title)), Ul(*sub_items, cls="ml-4 pl-2 space-y-1 border-l border-slate-100 dark:border-slate-800"), data_folder="true", open=should_expand), cls="my-1"))
            elif note_slug:
                items.append(Li(A(Span(cls="w-4 mr-2 shrink-0"), Span(UkIcon("folder", cls="text-blue-500 w-4 h-4"), cls="w-4 mr-2 flex items-center justify-center shrink-0"), Span(folder_title, cls="whitespace-nowrap", title=folder_title), href=f"/posts/{note_slug}", hx_get=f"/posts/{note_slug}", hx_target="#main-content", hx_push_url="true", hx_swap="outerHTML show:window:top settle:0.1s", cls="post-link inline-flex w-max items-center py-1 px-2 rounded hover:bg-slate-100 dark:hover:bg-slate-800 text-slate-700 dark:text-slate-300 hover:text-blue-600 hover:underline transition-colors whitespace-nowrap", data_path=note_slug)))
            continue
        if item.suffix not in {".md", ".pdf", ".excalidraw"}:
            continue
        slug = str(item.relative_to(root).with_suffix(""))
        route = f"/drawings/{slug}" if item.suffix == ".excalidraw" else f"/posts/{slug}"
        if not is_allowed_fn(route, roles or [], rbac_rules):
            continue
        if item.suffix == ".md":
            metadata, _ = parse_frontmatter_fn(item)
            icon = "monitor" if metadata.get("slides", False) else "file-text"
            title = metadata.get("title", slug_to_title_fn(item.stem, abbreviations=abbreviations))
            label, href = title, f"/posts/{slug}"
        elif item.suffix == ".pdf":
            icon, title, label, href = "file", slug_to_title_fn(item.stem, abbreviations=abbreviations), f"{slug_to_title_fn(item.stem, abbreviations=abbreviations)} (PDF)", f"/posts/{slug}"
        else:
            icon, title, label, href = "pencil", slug_to_title_fn(item.stem, abbreviations=abbreviations), f"{slug_to_title_fn(item.stem, abbreviations=abbreviations)} (Excalidraw)", f"/drawings/{slug}"
        items.append(Li(A(Span(cls="w-4 mr-2 shrink-0"), Span(UkIcon(icon, cls="text-slate-400 w-4 h-4"), cls="w-4 mr-2 flex items-center justify-center shrink-0"), Span(label, cls="whitespace-nowrap", title=title), href=href, hx_get=href, hx_target="#main-content", hx_push_url="true", hx_swap="outerHTML show:window:top settle:0.1s", cls="post-link inline-flex items-center py-1 px-2 rounded hover:bg-slate-100 dark:hover:bg-slate-800 text-slate-700 dark:text-slate-300 hover:text-blue-600 transition-colors whitespace-nowrap", data_path=slug)))
    logger.debug(f"[DEBUG] build_post_tree for {folder.relative_to(root) if folder != root else '.'} completed in {(time.time() - start_time) * 1000:.2f}ms")
    return items



import html
import json
import re
import time
from pathlib import Path
from urllib.parse import quote

from fasthtml.common import A, Button, Div, H1, Li, NotStr, Ol, P, Response, Script, Strong, Textarea
from monsterui.all import UkIcon


def render_post_detail(path, htmx, request, *, get_root_folder, effective_abbreviations, find_folder_note_file, slug_to_title, layout, get_blog_title, not_found, parse_frontmatter, resolve_markdown_title, from_md, logger, PathCls=Path):
    request_start = time.time()
    logger.info(f"\n[DEBUG] ########## REQUEST START: /posts/{path} ##########")
    root = get_root_folder()
    abbreviations = effective_abbreviations(root)
    file_path, pdf_path, excalidraw_path, folder_path = root / f"{path}.md", root / f"{path}.pdf", root / f"{path}.excalidraw", root / path
    if not file_path.exists():
        if folder_path.exists() and folder_path.is_dir():
            note_file = find_folder_note_file(folder_path)
            if note_file:
                from starlette.responses import RedirectResponse
                return RedirectResponse(f"/posts/{note_file.relative_to(root).with_suffix('')}", status_code=307)
        if pdf_path.exists():
            post_title = f"{slug_to_title(PathCls(path).name, abbreviations=abbreviations)} (PDF)"
            pdf_src = f"/posts/{path}.pdf"
            pdf_content = Div(Div(H1(post_title, cls="text-4xl font-bold"), Button("Focus PDF", cls="pdf-focus-toggle inline-flex items-center gap-2 px-3 py-1.5 text-sm font-medium rounded-md border border-slate-200 dark:border-slate-700 bg-white dark:bg-slate-900 hover:bg-slate-50 dark:hover:bg-slate-800 transition-colors", type="button", data_pdf_focus_toggle="true", data_pdf_focus_label="Focus PDF", data_pdf_exit_label="Exit focus", aria_pressed="false"), cls="flex items-center justify-between gap-4 flex-wrap mb-6"), NotStr(f'<object data="{pdf_src}" type="application/pdf" class="pdf-viewer w-full h-[calc(100vh-14rem)] rounded-lg border border-slate-200 dark:border-slate-700 bg-white dark:bg-slate-900"><p class="p-4 text-sm text-slate-600 dark:text-slate-300">PDF preview not available. <a href="{pdf_src}" class="text-blue-600 hover:underline">Download PDF</a>.</p></object>'))
            return layout(pdf_content, htmx=htmx, title=f"{post_title} - {get_blog_title()}", show_sidebar=True, toc_content=None, current_path=path, show_toc=False, auth=request.scope.get("auth"))
        return not_found(htmx, auth=request.scope.get("auth"))
    metadata, raw_content = parse_frontmatter(file_path)
    post_title, render_content = resolve_markdown_title(file_path, abbreviations=abbreviations)
    md_start = time.time()
    content = from_md(render_content, current_path=path)
    logger.debug(f"[DEBUG] Markdown rendering took {(time.time() - md_start) * 1000:.2f}ms")
    copy_button = Button(UkIcon("clipboard", cls="w-4 h-4"), type="button", title="Copy raw markdown", onclick="(function(){const el=document.getElementById('raw-md-clipboard');const toast=document.getElementById('raw-md-toast');if(!el){return;}el.focus();el.select();const text=el.value;const done=()=>{if(!toast){return;}toast.classList.remove('opacity-0');toast.classList.add('opacity-100');setTimeout(()=>{toast.classList.remove('opacity-100');toast.classList.add('opacity-0');},1400);};if(navigator.clipboard&&window.isSecureContext){navigator.clipboard.writeText(text).then(done).catch(()=>{document.execCommand('copy');done();});}else{document.execCommand('copy');done();}})()", cls="inline-flex items-center justify-center p-2 rounded-md border border-slate-200 dark:border-slate-700 text-slate-600 dark:text-slate-300 hover:text-slate-900 dark:hover:text-white hover:border-slate-300 dark:hover:border-slate-500 transition-colors")
    present_button = A(UkIcon("monitor", cls="w-4 h-4"), "Present", href=f"/slides/{path}", target="_blank", rel="noopener noreferrer", cls="inline-flex items-center gap-2 px-3 py-2 rounded-md border border-slate-200 dark:border-slate-700 text-slate-600 dark:text-slate-300 hover:text-slate-900 dark:hover:text-white hover:border-slate-300 dark:hover:border-slate-500 transition-colors text-sm") if bool(metadata.get("slides", False)) else None
    post_content = Div(Div(H1(post_title, cls="text-4xl font-bold"), present_button, copy_button, cls="flex items-center gap-2 flex-wrap mb-8"), Div("Copied Raw Markdown!", id="raw-md-toast", cls="fixed top-6 right-6 bg-slate-900 text-white text-sm px-4 py-2 rounded shadow-lg opacity-0 transition-opacity duration-300"), Textarea(raw_content, id="raw-md-clipboard", cls="absolute left-[-9999px] top-0 opacity-0 pointer-events-none"), content)
    layout_start = time.time()
    result = layout(post_content, htmx=htmx, title=f"{post_title} - {get_blog_title()}", show_sidebar=True, toc_content=raw_content, current_path=path, auth=request.scope.get("auth"))
    logger.debug(f"[DEBUG] Layout generation took {(time.time() - layout_start) * 1000:.2f}ms")
    logger.debug(f"[DEBUG] ########## REQUEST COMPLETE: {(time.time() - request_start) * 1000:.2f}ms TOTAL ##########\n")
    return result


def render_drawing_detail(path, htmx, request, *, get_root_folder, not_found, get_roles_from_request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list, is_allowed, slug_to_title, effective_abbreviations, drawing_password_for, get_blog_title, layout):
    root, file_path = get_root_folder(), get_root_folder() / f"{path}.excalidraw"
    if not file_path.exists():
        return not_found(htmx, auth=request.scope.get("auth"))
    if htmx and getattr(htmx, "request", None):
        return Response(status_code=200, headers={"HX-Redirect": f"/drawings/{path}"})
    roles = get_roles_from_request(request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list)
    if roles is not None and not is_allowed(f"/posts/{path}", roles or [], rbac_rules):
        return not_found(htmx, auth=request.scope.get("auth"))
    title = f"{slug_to_title(Path(path).name, abbreviations=effective_abbreviations(root))} (Excalidraw)"
    host_id, drawing_protected = f"excalidraw-{abs(hash(path)) & 0xFFFFFF}", bool(drawing_password_for(path))
    download_url, download_name, auth = f"/download/{path}.excalidraw", f"{Path(path).name}.excalidraw", request.scope.get("auth")
    default_user = (auth.get("name") or auth.get("email") or auth.get("username") or "") if auth else ""
    post_content = Div(Script("document.body.dataset.forceFullNav='1';"), Div(H1(title, cls="text-4xl font-bold"), Div(Button(default_user or "Set your name", type="button", data_excalidraw_name=host_id, data_excalidraw_name_locked="1" if auth else "0", data_excalidraw_name_default=default_user, disabled=bool(auth), cls="px-3 py-2 rounded-md border border-slate-200 dark:border-slate-700 text-sm " + ("opacity-70 cursor-default" if auth else "hover:bg-slate-100 dark:hover:bg-slate-800")), Button("Enable editing", type="button", data_excalidraw_toggle=host_id, cls="px-3 py-2 rounded-md border border-slate-200 dark:border-slate-700 text-sm hover:bg-slate-100 dark:hover:bg-slate-800"), A("Download .excalidraw", href=download_url, download=download_name, cls="px-3 py-2 rounded-md border border-slate-200 dark:border-slate-700 text-sm hover:bg-slate-100 dark:hover:bg-slate-800"), Button("Open Excalidraw", type="button", data_excalidraw_open_external="1", data_excalidraw_download_url=download_url, data_excalidraw_download_name=download_name, cls="px-3 py-2 rounded-md border border-slate-200 dark:border-slate-700 text-sm hover:bg-slate-100 dark:hover:bg-slate-800"), cls="flex items-center gap-3"), cls="flex items-center justify-between gap-3 mb-6 flex-wrap"), Div(id=host_id, cls="excalidraw-host w-full flex-1 min-h-0 rounded-lg border border-slate-200 dark:border-slate-700 overflow-hidden", data_excalidraw_path=path, data_excalidraw_src=f"/posts/{path}.excalidraw", data_excalidraw_save_url=f"/api/excalidraw/{path}", data_excalidraw_unlock_url=f"/api/excalidraw/unlock/{path}", data_excalidraw_protected="1" if drawing_protected else "0"), cls="h-[calc(100vh-8rem)] flex flex-col overflow-hidden")
    return layout(post_content, htmx=htmx, title=f"{title} - {get_blog_title()}", show_sidebar=True, toc_content=None, current_path=path, show_toc=False, auth=request.scope.get("auth"), nav_posts_menu=True, full_width=True, show_footer=False, no_scroll=True)


def render_slide_deck(path, request, *, get_root_folder, not_found, get_roles_from_auth, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list, is_allowed, parse_frontmatter, slug_to_title, effective_abbreviations, from_md, asset_url):
    root, file_path = get_root_folder(), get_root_folder() / f"{path}.md"
    if not file_path.exists():
        return not_found(auth=request.scope.get("auth"))
    roles = get_roles_from_auth(request.scope.get("auth"), rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list)
    if not is_allowed(f"/posts/{path}", roles or [], rbac_rules):
        return not_found(auth=request.scope.get("auth"))
    metadata, raw_content = parse_frontmatter(file_path)
    title = metadata.get("title", slug_to_title(Path(path).name, abbreviations=effective_abbreviations(root)))
    safe_title = html.escape(f"{title} - Slides")
    reveal_block = metadata.get("reveal", {}) if isinstance(metadata.get("reveal"), dict) else {}
    reveal_cfg = {**reveal_block, **{k[7:]: v for k, v in metadata.items() if k.startswith("reveal_")}}
    theme = str(reveal_cfg.pop("theme", "black")).strip() or "black"
    highlight_theme = str(reveal_cfg.pop("highlightTheme", "monokai")).strip() or "monokai"
    theme = theme if re.fullmatch(r"[a-zA-Z0-9_-]+", theme) else "black"
    highlight_theme = highlight_theme if re.fullmatch(r"[a-zA-Z0-9_-]+", highlight_theme) else "monokai"
    md_separator, md_separator_vertical = str(reveal_cfg.pop("separator", "^---$")), str(reveal_cfg.pop("separatorVertical", "^--$"))
    reveal_cfg.pop("separatorNotes", None)
    slide_padding = str(reveal_cfg.pop("slidePadding", "1.25rem")).strip() or "1.25rem"
    reveal_cfg.pop("margin", None)
    font_size = str(reveal_cfg.pop("fontSize", "18px")).strip() or "18px"
    right_advances_all = reveal_cfg.pop("rightAdvancesAll", False)
    if isinstance(right_advances_all, str):
        right_advances_all = right_advances_all.strip().lower() in {"1", "true", "yes", "on"}
    if right_advances_all and "navigationMode" not in reveal_cfg:
        reveal_cfg["navigationMode"] = "linear"
    if not re.fullmatch(r"[0-9]+(?:\.[0-9]+)?(?:px|rem|em|vw|vh|%)", font_size):
        font_size = "18px"
    reveal_init_json, sep_re, sep_v_re = json.dumps({"hash": True, "slideNumber": True, "margin": 0, **reveal_cfg}, ensure_ascii=False), re.compile(md_separator), re.compile(md_separator_vertical)

    def _split_slide_groups(md_text):
        groups, group, buf, in_fence, fence_char, fence_len = [], [], [], False, "", 0
        for line in md_text.splitlines():
            s = line.strip()
            m = re.match(r"^(```+|~~~+)", s)
            if m:
                tok = m.group(1)
                if not in_fence:
                    in_fence, fence_char, fence_len = True, tok[0], len(tok)
                elif tok[0] == fence_char and len(tok) >= fence_len:
                    in_fence = False
            if not in_fence and sep_re.fullmatch(s):
                group.append("\n".join(buf).strip()); groups.append([x for x in group if x.strip()]); group, buf = [], []; continue
            if not in_fence and sep_v_re.fullmatch(s):
                group.append("\n".join(buf).strip()); buf = []; continue
            buf.append(line)
        group.append("\n".join(buf).strip()); groups.append([x for x in group if x.strip()])
        return [g for g in groups if g]

    def _render_slide_fragment(md_fragment):
        from fasthtml.common import to_xml
        return re.sub(r"<link[^>]*sidenote\\.css[^>]*>", "", to_xml(from_md(md_fragment, current_path=path)), count=1)

    sections = []
    for group in _split_slide_groups(raw_content):
        sections.append(f"<section>{_render_slide_fragment(group[0])}</section>" if len(group) == 1 else f"<section>{''.join(f'<section>{_render_slide_fragment(item)}</section>' for item in group)}</section>")
    slides_markup = "".join(sections) if sections else "<section><h2>Empty deck</h2></section>"
    page_html = f"""<!doctype html><html><head><meta charset="utf-8"><meta name="viewport" content="width=device-width, initial-scale=1"><title>{safe_title}</title><link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/reveal.js@5/dist/reveal.css"><link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/reveal.js@5/dist/theme/{theme}.css" id="theme"><link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/reveal.js@5/plugin/highlight/{highlight_theme}.css"><link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/katex@0.16.9/dist/katex.min.css"><style>:root{{--vyasa-slide-padding:{slide_padding};--vyasa-slide-font-size:{font_size};}}.reveal{{font-size:var(--vyasa-slide-font-size);}}.reveal .slides section{{padding:var(--vyasa-slide-padding);}}.reveal pre code{{max-height:none;}}.reveal .slides section.present{{left:0!important;}}.reveal section img{{max-height:72vh;}}.reveal .mermaid-container,.reveal .d2-container{{position:relative;border:1px solid rgba(15,23,42,.18)!important;border-radius:10px!important;box-shadow:none!important;background:transparent!important;padding:14px!important;box-sizing:border-box!important;left:auto!important;transform:none!important;margin:0 auto!important;width:85%!important;max-width:85%!important;height:85%!important;max-height:85%!important;min-height:0!important;align-self:center!important;}}.reveal .mermaid-controls,.reveal .d2-controls{{display:none!important;}}.reveal .mermaid-wrapper,.reveal .d2-wrapper{{overflow:visible;min-height:0!important;height:100%!important;width:100%!important;justify-content:center!important;align-items:center!important;}}.reveal .mermaid-wrapper svg,.reveal .d2-wrapper svg{{width:100%!important;height:100%!important;max-width:100%!important;max-height:100%!important;}}.reveal .slides section:has(.mermaid-container),.reveal .slides section:has(.d2-container){{display:flex!important;flex-direction:column!important;align-items:center!important;justify-content:center!important;padding:0!important;width:100%!important;height:100%!important;min-height:100%!important;}}.reveal .slides section:has(.mermaid-container)>*,.reveal .slides section:has(.d2-container)>*{{width:100%!important;height:100%!important;min-width:0!important;min-height:0!important;display:flex!important;flex-direction:column!important;align-items:center!important;justify-content:center!important;}}.reveal .mermaid,.reveal .mermaid svg{{font-size:16px!important;line-height:1.2!important;}}.reveal .mermaid svg text,.reveal .mermaid svg tspan{{fill:#1f2937!important;}}.reveal .mermaid .nodeLabel,.reveal .mermaid .edgeLabel,.reveal .mermaid foreignObject div,.reveal .mermaid foreignObject span,.reveal .mermaid foreignObject p{{color:#1f2937!important;fill:#1f2937!important;}}.reveal .code-copy-button,.reveal .code-block [id$="-toast"]{{display:none!important;}}.reveal .code-block textarea{{position:absolute!important;left:-9999px!important;top:0!important;opacity:0!important;pointer-events:none!important;}}</style></head><body><div class="reveal"><div class="slides">{slides_markup}</div></div><script src="https://cdn.jsdelivr.net/npm/reveal.js@5/dist/reveal.js"></script><script src="https://cdn.jsdelivr.net/npm/reveal.js@5/plugin/highlight/highlight.js"></script><script src="https://cdn.jsdelivr.net/npm/katex@0.16.9/dist/katex.min.js"></script><script src="https://cdn.jsdelivr.net/npm/katex@0.16.9/dist/contrib/auto-render.min.js"></script><script type="module" src="{asset_url('/static/scripts.js')}"></script><script>const cfg = {reveal_init_json};cfg.plugins = [RevealHighlight];Reveal.initialize(cfg);function runMathPass(node) {{if (typeof renderMathInElement !== 'function') return;renderMathInElement(node || document.body, {{delimiters: [{{left: '$$', right: '$$', display: true}},{{left: '$', right: '$', display: false}},{{left: '\\\\[', right: '\\\\]', display: true}},{{left: '\\\\(', right: '\\\\)', display: false}}],throwOnError: false}});}}runMathPass(document.body);Reveal.on('ready', () => runMathPass(document.body));Reveal.on('slidechanged', (e) => runMathPass(e.currentSlide || document.body));</script></body></html>"""
    return Response(page_html, media_type="text/html; charset=utf-8")


def find_index_file(get_root_folder):
    root = get_root_folder()
    for file in root.iterdir():
        if file.is_file() and file.suffix == ".md" and file.stem.lower() == "index":
            return file
    for file in root.iterdir():
        if file.is_file() and file.suffix == ".md" and file.stem.lower() == "readme":
            return file
    return None


def render_index(htmx, request, *, get_blog_title, find_index_file_fn, parse_frontmatter, resolve_markdown_title, get_root_folder, from_md, layout, not_found=None, logger=None):
    request_start = time.time()
    if logger:
        logger.debug("Request start path=/ route=index")
    blog_title = get_blog_title()
    index_file = find_index_file_fn()
    if index_file:
        _, raw_content = parse_frontmatter(index_file)
        page_title, render_content = resolve_markdown_title(index_file)
        index_path = str(index_file.relative_to(get_root_folder()).with_suffix(""))
        page_content = Div(H1(page_title, cls="text-4xl font-bold mb-8"), from_md(render_content, current_path=index_path))
        result = layout(page_content, htmx=htmx, title=f"{page_title} - {blog_title}", show_sidebar=True, toc_content=raw_content, current_path=index_path, auth=request.scope.get("auth"))
        if logger:
            logger.debug("Request complete path=/ route=index total={:.2f}ms", (time.time() - request_start) * 1000)
        return result
    result = layout(Div(H1(f"Welcome to {blog_title}!", cls="text-4xl font-bold tracking-tight mb-8"), P("Quick start tutorial", cls="text-lg font-medium text-slate-700 dark:text-slate-300 mb-4"), Ol(Li("Use the sidebar to browse the files and folders in your blog."), Li("Open a markdown file to preview it instantly."), Li("Create an ", Strong("index.md"), " or ", Strong("README.md"), " in your blog directory to replace this page with your own landing page."), cls="list-decimal pl-6 space-y-2 text-base text-slate-600 dark:text-slate-400 mb-4"), P("More guides, examples, and documentation are available at ", A("vyasa.yeshwanth.com", href="https://vyasa.yeshwanth.com", cls="text-slate-900 dark:text-slate-100 underline underline-offset-4"), ".", cls="text-base text-slate-600 dark:text-slate-400"), cls="w-full"), htmx=htmx, title=f"Home - {blog_title}", show_sidebar=True, auth=request.scope.get("auth"))
    if logger:
        logger.debug("Request complete path=/ route=index total={:.2f}ms", (time.time() - request_start) * 1000)
    return result



from fasthtml.common import *


def login_content(error, google_enabled, local_enabled):
    return Div(
        H2("Login", cls="uk-h2"),
        (A(Span("Continue with Google", cls="text-sm font-semibold"), href="/login/google", cls="inline-flex items-center justify-center px-4 py-2 my-6 rounded-md border border-slate-700 bg-slate-800 text-slate-100 hover:bg-slate-900 hover:border-slate-900 dark:bg-slate-800/80 dark:text-slate-100 dark:hover:bg-slate-900/80 transition-colors max-w-sm mx-auto") if google_enabled else None),
        (Form(Div(Input(type="text", name="username", required=True, id="username", cls="uk-input input input-bordered w-full", placeholder="Username"), cls="my-4"), Div(Input(type="password", name="password", required=True, id="password", cls="uk-input input input-bordered w-full", placeholder="Password"), cls="my-4"), Button("Login", type="submit", cls="uk-btn btn btn-primary w-full"), enctype="multipart/form-data", method="post", cls="max-w-sm mx-auto") if local_enabled else None),
        P(error, cls="text-red-500 mt-4") if error else None,
        cls="prose mx-auto mt-24 text-center",
    )


def impersonate_content(error, success, impersonating_email):
    return Div(
        H1("Impersonate User", cls="text-3xl font-bold"),
        P("Switch the current session to a different user for RBAC testing.", cls="text-slate-600 dark:text-slate-400"),
        Div(P(error, cls="text-red-600") if error else None, P(success, cls="text-emerald-600") if success else None, cls="mt-4"),
        Div((P(f"Currently impersonating: {impersonating_email}", cls="text-sm text-amber-600 dark:text-amber-400") if impersonating_email else None), cls="mt-2"),
        Form(
            Div(Label("User email", cls="block text-sm font-medium mb-2"), Input(type="email", name="email", placeholder="user@domain.com", cls="w-full px-3 py-2 rounded-md border border-slate-200 dark:border-slate-700 bg-white/80 dark:bg-slate-900/60"), cls="mt-6"),
            Div(Button("Start Impersonation", type="submit", name="action", value="start", cls="mt-6 px-4 py-2 rounded-md bg-blue-600 text-white hover:bg-blue-700"), Button("Stop Impersonation", type="submit", name="action", value="stop", cls="mt-6 ml-3 px-4 py-2 rounded-md bg-slate-200 text-slate-900 hover:bg-slate-300 dark:bg-slate-700 dark:text-slate-100 dark:hover:bg-slate-600"), cls="flex items-center"),
            method="post",
            cls="mt-4",
        ),
        cls="max-w-xl mx-auto py-10 px-6",
    )



import re
import mistletoe as mst


def span_token(name, pat, attr, prec=5):
    class T(mst.span_token.SpanToken):
        precedence, parse_inner, parse_group, pattern = prec, False, 1, re.compile(pat)
        def __init__(self, match):
            setattr(self, attr, match.group(1))
            if hasattr(match, "lastindex") and match.lastindex and match.lastindex >= 2:
                if name == "YoutubeEmbed": self.caption = match.group(2) if match.group(2) else None
                elif name == "MermaidEmbed": self.option = match.group(2) if match.group(2) else None
                elif name == "IframeEmbed": self.options = match.group(2) if match.group(2) else None
                elif name == "DownloadEmbed": self.label = match.group(2) if match.group(2) else None
    T.__name__ = name
    return T


FootnoteRef = span_token("FootnoteRef", r"\[\^([^\]]+)\](?!:)", "target")
YoutubeEmbed = span_token("YoutubeEmbed", r"\[yt:([a-zA-Z0-9_-]+)(?:\|(.+))?\]", "video_id", 6)
IframeEmbed = span_token("IframeEmbed", r"\[iframe:([^\|\]]+)(?:\|(.+))?\]", "src", 6)
DownloadEmbed = span_token("DownloadEmbed", r"\[download:([^\|\]]+)(?:\|(.+))?\]", "path", 6)


class Superscript(mst.span_token.SpanToken):
    pattern = re.compile(r"(?<![\\\w$])\^([A-Za-z0-9.+\-]{1,32})\^(?![\w$])"); parse_inner = False; parse_group = 1; precedence = 7
    def __init__(self, match): self.content = match.group(1); self.children = []


class Subscript(mst.span_token.SpanToken):
    pattern = re.compile(r"(?<![~\\\w$])~([A-Za-z0-9.+\-]{1,32})~(?![~\w$])"); parse_inner = False; parse_group = 1; precedence = 7
    def __init__(self, match): self.content = match.group(1); self.children = []


class InlineCodeAttr(mst.span_token.SpanToken):
    pattern = re.compile(r"`([^`]+)`\{([^\}]+)\}"); parse_inner = False; parse_group = 1; precedence = 8
    def __init__(self, match): self.code = match.group(1); self.attrs = match.group(2); self.children = []


class Strikethrough(mst.span_token.SpanToken):
    pattern = re.compile(r"~~(.+?)~~"); parse_inner = True; parse_group = 1; precedence = 7
    def __init__(self, match): self.children = []



import re
from functools import lru_cache


def normalize_search_text(text):
    text = (text or "").lower().replace("-", " ").replace("_", " ")
    return " ".join(text.split())


def parse_search_query(query):
    trimmed = (query or "").strip()
    if len(trimmed) >= 2 and trimmed.startswith("/") and trimmed.endswith("/"):
        pattern = trimmed[1:-1].strip()
        if not pattern:
            return None, ""
        try:
            return re.compile(pattern, re.IGNORECASE), ""
        except re.error:
            return None, "Invalid regex. Showing normal matches instead."
    return None, ""


@lru_cache(maxsize=256)
def cached_search_matches(fingerprint, show_hidden, query, limit, finder):
    return finder(query, limit)


def find_search_matches(query, limit, fingerprint, show_hidden, finder):
    return cached_search_matches(fingerprint, show_hidden, query, limit, finder)



from pathlib import Path

from fasthtml.common import Beforeware
from starlette.staticfiles import StaticFiles

AUTH_SKIP_ROUTES = [
    r"^/login$",
    r"^/login/google$",
    r"^/auth/google/callback$",
    r"^/_sidebar/.*",
    r"^/static/.*",
    r"^/chat/.*",
    r".*\.css",
    r".*\.js",
]


def build_beforeware(handler, enabled: bool):
    if not enabled:
        return None
    return Beforeware(handler, skip=AUTH_SKIP_ROUTES)


def build_app(app_factory, hdrs, beforeware):
    return app_factory(hdrs=hdrs, before=beforeware, exts="ws") if beforeware else app_factory(hdrs=hdrs, exts="ws")


def mount_package_static(app_instance, package_dir: Path):
    static_dir = package_dir / "static"
    if static_dir.exists():
        app_instance.mount("/static", StaticFiles(directory=str(static_dir)), name="static")



import asyncio
import json


class CollabRuntime:
    def __init__(self):
        self.rooms = {}
        self.scenes = {}
        self.lock = asyncio.Lock()

    async def broadcast(self, room, payload, exclude=None):
        msg = json.dumps(payload)
        for sock in list(self.rooms.get(room, set())):
            if exclude is not None and sock is exclude:
                continue
            try:
                await sock.send_text(msg)
            except Exception:
                pass

    async def connect(self, ws):
        room = ws.path_params.get("path", "")
        async with self.lock:
            self.rooms.setdefault(room, set()).add(ws)
            scene = self.scenes.get(room)
        if scene is not None:
            await ws.send_text(json.dumps({"type": "scene", "room": room, "scene": scene}))

    async def disconnect(self, ws):
        room = ws.path_params.get("path", "")
        peer_id = ws.scope.get("collab_id")
        async with self.lock:
            peers = self.rooms.get(room, set())
            peers.discard(ws)
            if not peers:
                self.rooms.pop(room, None)
        if peer_id:
            await self.broadcast(room, {"type": "presence_remove", "id": peer_id}, exclude=ws)



from pathlib import Path


def load_pylogue_routes(logger):
    try:
        from pylogue.core import register_routes, EchoResponder
        return register_routes, EchoResponder
    except Exception:
        pylogue_path = Path("/Users/yeshwanth/Code/Personal/pylogue/src/pylogue/core.py")
        if not pylogue_path.exists():
            logger.warning(f"Pylogue not found at {pylogue_path}")
            return None, None
        try:
            import importlib.util
            spec = importlib.util.spec_from_file_location("pylogue.core", pylogue_path)
            if spec and spec.loader:
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                return module.register_routes, module.EchoResponder
        except Exception as exc:
            logger.warning(f"Failed to load pylogue from {pylogue_path}: {exc}")
            return None, None
    return None, None


def register_chat_routes(app_instance, rt, logger, title):
    pylogue_register, pylogue_responder = load_pylogue_routes(logger)
    if not pylogue_register:
        return
    try:
        from .agent import PydanticAIStreamingResponder
        chat_responder_factory = PydanticAIStreamingResponder
        logger.info("Using PydanticAIStreamingResponder for /chat")
    except Exception as exc:
        logger.warning(f"Falling back to Pylogue responder: {exc}")
        chat_responder_factory = pylogue_responder
    pylogue_register(app_instance, responder_factory=chat_responder_factory, title=title, subtitle="Ask a question about this blog", tag_line="« Blog", tag_line_href="/", base_path="chat", inject_headers=True)

    @rt("/chat")
    def chat_redirect():
        from starlette.responses import RedirectResponse
        return RedirectResponse("/chat/", status_code=307)



import html
import re
from functools import partial
from itertools import count
from pathlib import Path

import mistletoe as mst
from fasthtml.common import Div, Link, NotStr, Span, to_xml
from loguru import logger
from monsterui.all import UkIcon, apply_classes

from .assets import asset_url
from .config import get_config
from .helpers import (
    _plain_text_from_html,
    _unique_anchor,
    parse_frontmatter,
    resolve_heading_anchor,
    text_to_anchor,
)
from .markdown_pipeline import (
    extract_footnotes,
    preprocess_callouts,
    preprocess_code_includes,
    preserve_newlines as preserve_md_newlines,
    preprocess_super_sub,
)
from .markdown_tabs import postprocess_tabs as postprocess_md_tabs
from .markdown_tabs import preprocess_tabs as preprocess_md_tabs
from .markdown_tokens import (
    DownloadEmbed,
    FootnoteRef,
    IframeEmbed,
    InlineCodeAttr,
    Strikethrough,
    Subscript,
    Superscript,
    YoutubeEmbed,
)

_diagram_uid_counter = count(1)
_CALLOUT_META = {
    "note": ("Note", "info"), "abstract": ("Abstract", "file-text"), "info": ("Info", "info"),
    "todo": ("Todo", "check"), "tip": ("Tip", "bolt"), "success": ("Success", "check"),
    "question": ("Question", "question"), "warning": ("Warning", "warning"), "failure": ("Failure", "close"),
    "danger": ("Danger", "warning"), "bug": ("Bug", "bug"), "example": ("Example", "code"), "quote": ("Quote", "quote-right"),
}
_CALLOUT_SVGS = {
    "info": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><circle cx="12" cy="12" r="9"/><path d="M12 10v6"/><path d="M12 7h.01"/></svg>',
    "file-text": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M14 3H7a2 2 0 0 0-2 2v14a2 2 0 0 0 2 2h10a2 2 0 0 0 2-2V8z"/><path d="M14 3v5h5"/><path d="M9 13h6"/><path d="M9 17h6"/><path d="M9 9h1"/></svg>',
    "check": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="m5 12 5 5L20 7"/></svg>',
    "bolt": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M13 2 4 14h6l-1 8 9-12h-6z"/></svg>',
    "question": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><circle cx="12" cy="12" r="9"/><path d="M9.5 9a2.5 2.5 0 1 1 4.2 1.8c-.9.8-1.7 1.3-1.7 2.7"/><path d="M12 17h.01"/></svg>',
    "warning": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M12 3 2.8 19a1 1 0 0 0 .9 1.5h16.6a1 1 0 0 0 .9-1.5z"/><path d="M12 9v4"/><path d="M12 17h.01"/></svg>',
    "close": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><circle cx="12" cy="12" r="9"/><path d="m15 9-6 6"/><path d="m9 9 6 6"/></svg>',
    "bug": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M8 8a4 4 0 1 1 8 0"/><path d="M9 9h6v8a3 3 0 0 1-6 0z"/><path d="M3 13h4"/><path d="M17 13h4"/><path d="m5 7 3 2"/><path d="m19 7-3 2"/><path d="m5 19 3-2"/><path d="m19 19-3-2"/></svg>',
    "code": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="m8 16-4-4 4-4"/><path d="m16 8 4 4-4 4"/><path d="m14 4-4 16"/></svg>',
    "quote-right": '<svg viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2"><path d="M10 11H6a2 2 0 0 0-2 2v1a4 4 0 0 0 4 4h1"/><path d="M20 11h-4a2 2 0 0 0-2 2v1a4 4 0 0 0 4 4h1"/></svg>',
}


def _callout_label(kind):
    return _CALLOUT_META.get(kind, (kind.replace("-", " ").replace("_", " ").title(), "info"))[0]


def _callout_icon(kind):
    return _CALLOUT_SVGS.get(_CALLOUT_META.get(kind, ("", "info"))[1], _CALLOUT_SVGS["info"])


def _render_callout(kind, body, render_body, title=None, fold=None):
    theme_kind = kind if kind in _CALLOUT_META else "note"
    rendered = render_body(body).strip()
    heading = html.escape(title or _callout_label(kind))
    head_cls = "vyasa-callout-head vyasa-callout-head-with-body flex items-center gap-2" if rendered else "vyasa-callout-head flex items-center gap-2"
    chevron = '<span class="vyasa-callout-chevron" aria-hidden="true"></span>' if fold else ""
    head = f'<div class="{head_cls}"><span class="vyasa-callout-icon">{_callout_icon(theme_kind)}</span><span class="vyasa-callout-label text-sm font-semibold tracking-[0.02em]">{heading}</span>{chevron}</div>'
    body_html = f'<div class="vyasa-callout-body">{rendered}</div>' if rendered else ""
    if fold:
        open_attr = " open" if fold == "+" else ""
        return f'<details class="vyasa-callout vyasa-callout-{theme_kind} my-6 rounded-xl border px-5 py-4" data-callout="{html.escape(kind)}"{open_attr}><summary class="vyasa-callout-summary list-none cursor-pointer">{head}</summary>{body_html}</details>'
    return f'<div class="vyasa-callout vyasa-callout-{theme_kind} my-6 rounded-xl border px-5 py-4" data-callout="{html.escape(kind)}">{head}{body_html}</div>'


def _render_markdown_fragment(body, img_dir=None, current_path=None):
    rendered = to_xml(from_md(body, img_dir=img_dir, current_path=current_path))
    rendered = re.sub(r'^<div class="w-full">\s*<link[^>]+>\s*', "", rendered)
    return re.sub(r"\s*</div>\s*$", "", rendered)


def _infer_code_language(path):
    mapping = {
        ".py": "python", ".js": "javascript", ".ts": "typescript", ".tsx": "tsx",
        ".jsx": "jsx", ".json": "json", ".toml": "toml", ".md": "markdown",
        ".sh": "bash", ".html": "html", ".css": "css", ".yml": "yaml", ".yaml": "yaml",
    }
    return mapping.get(Path(path).suffix.lower(), "")


def _parse_line_spec(spec):
    match = re.search(r"ln\[(\d+):(\d+)\]", spec)
    return (int(match.group(1)), int(match.group(2))) if match else None


def _parse_highlight_spec(spec):
    match = re.search(r"hl\[([^\]]+)\]", spec)
    return match.group(1).replace(":", "-").replace(" ", "") if match else ""


def _render_code_include(snippet, lang="", start=1, highlight_spec=""):
    attrs = [f'data-code-source-start="{start}"']
    if highlight_spec:
        attrs.append(f'data-code-highlight-lines="{html.escape(highlight_spec)}"')
    lang_class = f' class="language-{lang}"' if lang else ""
    return (
        '<div class="code-block relative my-4">'
        f'<pre><code{lang_class} {" ".join(attrs)}>{html.escape(snippet)}</code></pre>'
        '</div>'
    )


class FrankenRenderer(mst.HTMLRenderer):
    def __init__(self, *args, img_dir=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.img_dir = img_dir

    def render_image(self, token):
        tpl = '<img src="{}" alt="{}"{}  class="max-w-full h-auto rounded-lg mb-6">'
        title = f' title="{token.title}"' if hasattr(token, "title") else ""
        src = token.src
        if self.img_dir and not src.startswith(
            ("http://", "https://", "/", "attachment:", "blob:", "data:")
        ):
            src = f"{self.img_dir}/{src}"
        return tpl.format(src, token.children[0].content if token.children else "", title)


def _asset_url(path):
    return asset_url(path)


def get_root_folder():
    return get_config().get_root_folder()


def _escape_attr(value):
    if value is None:
        return None
    return (
        str(value)
        .replace("&", "&amp;")
        .replace('"', "&quot;")
        .replace("'", "&#39;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
    )


def _render_d2_block(code):
    frontmatter_pattern = r"^---\s*\n(.*?)\n---\s*\n"
    frontmatter_match = re.match(frontmatter_pattern, code, re.DOTALL)
    height, width, min_height = "auto", "65vw", "320px"
    d2_layout, d2_theme_id, d2_dark_theme_id = "elk", None, None
    d2_sketch = d2_pad = d2_scale = d2_target = d2_animate_interval = d2_animate = d2_fullscreen_title = None
    if frontmatter_match:
        frontmatter_content = frontmatter_match.group(1)
        code = code[frontmatter_match.end():]
        try:
            config = dict(line.split(":", 1) for line in frontmatter_content.strip().split("\n") if ":" in line)
            config = {k.strip(): v.strip() for k, v in config.items()}
            height = config.get("height", height)
            min_height = height if "height" in config else min_height
            width = config.get("width", width)
            d2_layout = config.get("layout", d2_layout)
            d2_theme_id = config.get("theme_id")
            d2_dark_theme_id = config.get("dark_theme_id")
            d2_sketch = config.get("sketch")
            d2_pad = config.get("pad")
            d2_scale = config.get("scale")
            d2_target = config.get("target")
            d2_animate_interval = config.get("animate_interval", config.get("animate-interval"))
            d2_animate = config.get("animate")
            d2_fullscreen_title = config.get("title", config.get("fullscreen_title"))
        except Exception as e:
            print(f"Error parsing d2 frontmatter: {e}")
    diagram_id = f"d2-{abs(hash(code)) & 0xFFFFFF}-{next(_diagram_uid_counter)}"
    container_style = f"width: {width}; position: relative; left: 50%; transform: translateX(-50%);" if "vw" in str(width).lower() else f"width: {width};"
    escaped_code = code.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace('"', "&quot;").replace("'", "&#39;")
    pairs = [("layout", d2_layout), ("theme-id", d2_theme_id), ("dark-theme-id", d2_dark_theme_id), ("sketch", d2_sketch), ("pad", d2_pad), ("scale", d2_scale), ("target", d2_target), ("animate-interval", d2_animate_interval), ("animate", d2_animate), ("fullscreen-title", d2_fullscreen_title)]
    d2_attr_str = "".join(f' data-d2-{k}="{_escape_attr(v)}"' for k, v in pairs if v is not None)
    caption_html = f'<div class="text-xs text-slate-500 dark:text-slate-400 text-center px-3 pb-2">{html.escape(str(d2_fullscreen_title))}</div>' if d2_fullscreen_title else ""
    return f"""<div class="d2-container relative border-4 rounded-md my-4 shadow-2xl" style="{container_style}"><div class="d2-controls absolute top-2 right-2 z-10 flex gap-1 bg-white/80 dark:bg-slate-800/80 backdrop-blur-sm rounded"><button onclick="openD2Fullscreen('{diagram_id}')" class="px-2 py-1 text-xs border rounded hover:bg-slate-100 dark:hover:bg-slate-700" title="Fullscreen">⛶</button><button onclick="resetD2Zoom('{diagram_id}')" class="px-2 py-1 text-xs border rounded hover:bg-slate-100 dark:hover:bg-slate-700" title="Reset zoom">Reset</button><button onclick="zoomD2In('{diagram_id}')" class="px-2 py-1 text-xs border rounded hover:bg-slate-100 dark:hover:bg-slate-700" title="Zoom in">+</button><button onclick="zoomD2Out('{diagram_id}')" class="px-2 py-1 text-xs border rounded hover:bg-slate-100 dark:hover:bg-slate-700" title="Zoom out">−</button></div><div id="{diagram_id}" class="d2-wrapper p-4 overflow-hidden flex justify-center items-center" style="min-height: {min_height}; height: {height};" data-d2-code="{escaped_code}"{d2_attr_str}><pre class="d2" style="width: 100%; height: 100%; display: flex; align-items: center; justify-content: center;">{code}</pre></div>{caption_html}</div>"""


def _render_mermaid_block(code):
    frontmatter_pattern = r"^---\s*\n(.*?)\n---\s*\n"
    frontmatter_match = re.match(frontmatter_pattern, code, re.DOTALL)
    height, width, min_height, gantt_width, mermaid_title = "auto", "65vw", "400px", None, None
    if frontmatter_match:
        frontmatter_content = frontmatter_match.group(1)
        code = code[frontmatter_match.end():]
        try:
            config = dict(line.split(":", 1) for line in frontmatter_content.strip().split("\n") if ":" in line)
            config = {k.strip(): v.strip() for k, v in config.items()}
            height = config.get("height", height)
            min_height = height if "height" in config else min_height
            width = config.get("width", width)
            mermaid_title = config.get("title")
            if "aspect_ratio" in config:
                aspect_value = config["aspect_ratio"].strip()
                ratio = (lambda p: float(p[0]) / float(p[1]))(aspect_value.split(":")) if ":" in aspect_value else float(aspect_value)
                gantt_width = int(1200 * ratio)
        except Exception as e:
            print(f"Error parsing mermaid frontmatter: {e}")
    diagram_id = f"mermaid-{abs(hash(code)) & 0xFFFFFF}-{next(_diagram_uid_counter)}"
    container_style = f"width: {width}; position: relative; left: 50%; transform: translateX(-50%);" if "vw" in str(width).lower() else f"width: {width};"
    escaped_code = code.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;").replace('"', "&quot;").replace("'", "&#39;")
    gantt_data_attr = f' data-gantt-width="{gantt_width}"' if gantt_width else ""
    mermaid_title_attr = f' data-mermaid-title="{html.escape(str(mermaid_title))}"' if mermaid_title else ""
    caption_html = f'<div class="text-xs text-slate-500 dark:text-slate-400 text-center px-3 pb-2">{html.escape(str(mermaid_title))}</div>' if mermaid_title else ""
    return f"""<div class="mermaid-container relative border-4 rounded-md my-4 shadow-2xl" style="{container_style}"><div class="mermaid-controls absolute top-2 right-2 z-10 flex gap-1 bg-white/80 dark:bg-slate-800/80 backdrop-blur-sm rounded"><button onclick="openMermaidFullscreen('{diagram_id}')" class="px-2 py-1 text-xs border rounded hover:bg-slate-100 dark:hover:bg-slate-700" title="Fullscreen">⛶</button><button onclick="resetMermaidZoom('{diagram_id}')" class="px-2 py-1 text-xs border rounded hover:bg-slate-100 dark:hover:bg-slate-700" title="Reset zoom">Reset</button><button onclick="zoomMermaidIn('{diagram_id}')" class="px-2 py-1 text-xs border rounded hover:bg-slate-100 dark:hover:bg-slate-700" title="Zoom in">+</button><button onclick="zoomMermaidOut('{diagram_id}')" class="px-2 py-1 text-xs border rounded hover:bg-slate-100 dark:hover:bg-slate-700" title="Zoom out">−</button></div><div id="{diagram_id}" class="mermaid-wrapper p-4 overflow-hidden flex justify-center items-center" style="min-height: {min_height}; height: {height};" data-mermaid-code="{escaped_code}"{gantt_data_attr}{mermaid_title_attr}><pre class="mermaid" style="width: 100%; height: 100%; display: flex; align-items: center; justify-content: center;">{code}</pre></div>{caption_html}</div>"""


class ContentRenderer(FrankenRenderer):
    def __init__(self, *extras, img_dir=None, footnotes=None, current_path=None, **kwargs):
        super().__init__(*extras, img_dir=img_dir, **kwargs)
        self.footnotes, self.fn_counter = footnotes or {}, 0
        self.current_path = current_path
        self.heading_counts = {}
        self.mermaid_counter = 0
        self.iframe_counter = 0

    def render_list_item(self, token):
        inner = self.render_inner(token)
        task_pattern = re.match(r"^\s*\[([ xX])\]\s*(.*?)$", inner, re.DOTALL)
        if not task_pattern:
            task_pattern = re.match(r"^<p>\s*\[([ xX])\]\s*(.*?)</p>$", inner, re.DOTALL)
        if task_pattern:
            checked = task_pattern.group(1).lower() == "x"
            content = task_pattern.group(2).strip()
            if checked:
                checkbox_style = "background-color: #10b981; border-color: #10b981;"
                checkmark = '<svg class="w-full h-full text-white" viewBox="0 0 16 16" fill="none" stroke="currentColor" stroke-width="2.5"><polyline points="3,8 6,11 13,4"></polyline></svg>'
            else:
                checkbox_style = "background-color: #6b7280; border-color: #6b7280;"
                checkmark = ""
            checkbox = f"""<span class="inline-flex items-center justify-center mr-3 mt-0.5" style="width: 20px; height: 20px; border-radius: 6px; border: 2px solid; {checkbox_style} flex-shrink: 0;">
                {checkmark}
            </span>"""
            return f'<li class="task-list-item flex items-start" style="list-style: none; margin: 0.5rem 0;">{checkbox}<span class="flex-1">{content}</span></li>\n'
        return f"<li>{inner}</li>\n"

    def render_youtube_embed(self, token):
        video_id = token.video_id
        caption = getattr(token, "caption", None)
        iframe = f"""
        <div class="relative w-full aspect-video my-6 rounded-lg overflow-hidden border border-slate-200 dark:border-slate-800">
            <iframe
                src="https://www.youtube.com/embed/{video_id}"
                title="YouTube video"
                frameborder="0"
                allow="accelerometer; autoplay; clipboard-write; encrypted-media; gyroscope; picture-in-picture"
                allowfullscreen
                class="absolute inset-0 w-full h-full">
            </iframe>
        </div>
        """
        if caption:
            return iframe + f'<p class="text-sm text-slate-500 dark:text-slate-400 text-center mt-2">{caption}</p>'
        return iframe

    def render_iframe_embed(self, token):
        src = token.src.strip()
        options_raw = getattr(token, "options", None)
        width = "65vw"
        height = "400px"
        title = "Embedded content"
        allow = "clipboard-read; clipboard-write; fullscreen"
        allowfullscreen = True
        caption = None
        popup = False
        border = "default"
        if options_raw:
            for part in options_raw.split(";"):
                if not part.strip() or "=" not in part:
                    continue
                key, value = part.split("=", 1)
                key = key.strip().lower()
                value = value.strip()
                if key == "width":
                    width = value
                elif key == "height":
                    height = value
                elif key == "title":
                    title = value
                elif key == "allow":
                    allow = value
                elif key == "fullscreen":
                    allowfullscreen = value.lower() in ("1", "true", "yes", "on")
                elif key == "caption":
                    caption = value
                elif key == "popup":
                    popup = value.lower() in ("1", "true", "yes", "on")
                elif key == "border":
                    border = value.lower()
        break_out = "vw" in str(width).lower()
        container_style = f"width: {width}; position: relative; left: 50%; transform: translateX(-50%);" if break_out else f"width: {width};"
        self.iframe_counter += 1
        iframe_id = f"iframe-{abs(hash(src)) & 0xFFFFFF}-{self.iframe_counter}"
        fullscreen_button = ""
        if popup:
            fullscreen_button = (
                '<div class="iframe-controls absolute top-2 right-2 z-10 flex gap-1 '
                'bg-white/80 dark:bg-slate-800/80 backdrop-blur-sm rounded">'
                f'<button data-iframe-fullscreen-toggle="true" data-iframe-src="{src}" '
                f'data-iframe-title="{title}" data-iframe-allow="{allow}" '
                f'data-iframe-allowfullscreen="{str(allowfullscreen).lower()}" '
                'class="px-2 py-1 text-xs border rounded hover:bg-slate-100 '
                'dark:hover:bg-slate-700" title="Fullscreen">⛶</button></div>'
            )
        if border in ("black", "dark"):
            border_classes = "border border-black"
        elif border in ("none", "false", "0", "off"):
            border_classes = "border border-transparent"
        else:
            border_classes = "border border-slate-200 dark:border-slate-800"
        iframe = f"""
        <div class="relative my-6 rounded-lg overflow-hidden {border_classes}" style="{container_style}">
            {fullscreen_button}
            <iframe
                id="{iframe_id}"
                src="{src}"
                title="{title}"
                frameborder="0"
                allow="{allow}"
                {'allowfullscreen' if allowfullscreen else ''}
                style="width: 100%; height: {height};">
            </iframe>
        </div>
        """
        if caption:
            return iframe + f'<p class="text-sm text-slate-500 dark:text-slate-400 text-center mt-2">{caption}</p>'
        return iframe

    def render_download_embed(self, token):
        raw_path = token.path.strip()
        label = getattr(token, "label", None)
        if self.current_path:
            root = get_root_folder().resolve()
            current_file_full = root / self.current_path
            current_dir = current_file_full.parent
            resolved = (current_dir / raw_path).resolve()
            try:
                rel_path = resolved.relative_to(root).as_posix()
                download_path = f"/download/{rel_path}"
            except ValueError:
                download_path = raw_path
        else:
            download_path = f"/download/{raw_path}"
        if not label:
            label = Path(raw_path).name
        link_class = "text-amber-600 dark:text-amber-400 underline underline-offset-2 hover:text-amber-800 dark:hover:text-amber-200 font-medium transition-colors"
        return f'<a href="{download_path}" class="{link_class}" download hx-boost="false">{label}</a>'

    def render_footnote_ref(self, token):
        self.fn_counter += 1
        n, target = self.fn_counter, token.target
        content = self.footnotes.get(target, f"[Missing footnote: {target}]")
        if "\n" in content:
            content = content.replace("\r\n", "\n")
            placeholder = "__VYASA_PARA_BREAK__"
            content = content.replace("\n\n", f"\n{placeholder}\n")
            content = content.replace("\n", "<br>\n")
            content = content.replace(f"\n{placeholder}\n", "\n\n")
        rendered = mst.markdown(content, partial(ContentRenderer, img_dir=self.img_dir, current_path=self.current_path)).strip()
        if rendered.startswith("<p>") and rendered.endswith("</p>"):
            rendered = rendered[3:-4]
        style = "text-sm leading-relaxed border-l-2 border-amber-400 dark:border-blue-400 pl-3 text-neutral-500 dark:text-neutral-400 transition-all duration-500 w-full my-2 xl:my-0"
        toggle = f"on click if window.innerWidth >= 1280 then add .hl to #sn-{n} then wait 1s then remove .hl from #sn-{n} else toggle .open on me then toggle .show on #sn-{n}"
        ref = Span(id=f"snref-{n}", role="doc-noteref", aria_label=f"Sidenote {n}", cls="sidenote-ref cursor-pointer", _=toggle)
        note = Span(NotStr(rendered), id=f"sn-{n}", role="doc-footnote", aria_labelledby=f"snref-{n}", cls=f"sidenote {style}")
        hide = lambda c: to_xml(Span(c, cls="hidden", aria_hidden="true"))
        return hide(" (") + to_xml(ref) + to_xml(note) + hide(")")

    def render_heading(self, token):
        level = token.level
        inner = self.render_inner(token)
        plain = _plain_text_from_html(inner)
        heading_text, anchor = resolve_heading_anchor(plain, self.heading_counts)
        permalink = (
            f'<a href="#{anchor}" class="vyasa-heading-permalink ml-2 no-underline '
            f'text-slate-400 hover:text-slate-700 dark:hover:text-slate-200" '
            f'aria-label="Link to {html.escape(heading_text)}">¶</a>'
        )
        return f'<h{level} id="{anchor}"><span class="vyasa-heading-text">{html.escape(heading_text)}</span>{permalink}</h{level}>'

    def render_superscript(self, token):
        return f"<sup>{token.content}</sup>"

    def render_subscript(self, token):
        return f"<sub>{token.content}</sub>"

    def render_strikethrough(self, token):
        return f"<del>{self.render_inner(token)}</del>"

    def render_inline_code_attr(self, token):
        code = html.escape(token.code)
        attrs = token.attrs.strip()
        classes = []
        id_attr = None
        other_attrs = []
        for attr in re.findall(r"\.([^\s\.#]+)|#([^\s\.#]+)|([^\s\.#=]+)=([^\s\.#]+)", attrs):
            if attr[0]:
                classes.append(attr[0])
            elif attr[1]:
                id_attr = attr[1]
            elif attr[2]:
                other_attrs.append(f'{attr[2]}="{attr[3]}"')
        html_attrs = []
        if classes:
            html_attrs.append(f'class="{" ".join(classes)}"')
        if id_attr:
            html_attrs.append(f'id="{id_attr}"')
        html_attrs.extend(other_attrs)
        attr_str = " " + " ".join(html_attrs) if html_attrs else ""
        return f"<span{attr_str}>{code}</span>"

    def render_block_code(self, token):
        lang = getattr(token, "language", "")
        code = self.render_raw_text(token)
        if lang == "d2":
            return _render_d2_block(code)
        if lang == "mermaid":
            return _render_mermaid_block(code)
        raw_code = code
        code = html.unescape(code)
        if lang and lang.lower() != "markdown":
            code = html.escape(code)
        lang_class = f' class="language-{lang}"' if lang else ""
        icon_html = to_xml(UkIcon("copy", cls="w-4 h-4"))
        code_id = f"codeblock-{abs(hash(raw_code)) & 0xFFFFFF}"
        toast_id = f"{code_id}-toast"
        textarea_id = f"{code_id}-clipboard"
        escaped_raw = html.escape(raw_code)
        return (
            '<div class="code-block relative my-4">'
            f'<button type="button" class="code-copy-button absolute top-2 right-2 '
            "inline-flex items-center justify-center rounded border border-slate-200 "
            "dark:border-slate-700 bg-white/80 dark:bg-slate-900/70 "
            "text-slate-600 dark:text-slate-300 hover:text-slate-900 "
            "dark:hover:text-white hover:border-slate-300 dark:hover:border-slate-500 "
            f'transition-colors" aria-label="Copy code" onclick="(function(){{const el=document.getElementById(\'{textarea_id}\');const toast=document.getElementById(\'{toast_id}\');if(!el){{return;}}el.focus();el.select();const text=el.value;const done=()=>{{if(!toast){{return;}}toast.classList.remove(\'opacity-0\');toast.classList.add(\'opacity-100\');setTimeout(()=>{{toast.classList.remove(\'opacity-100\');toast.classList.add(\'opacity-0\');}},1400);}};if(navigator.clipboard&&window.isSecureContext){{navigator.clipboard.writeText(text).then(done).catch(()=>{{document.execCommand(\'copy\');done();}});}}else{{document.execCommand(\'copy\');done();}}}})()">'
            f'{icon_html}<span class="sr-only">Copy code</span></button>'
            f'<div id="{toast_id}" class="absolute top-2 right-10 text-xs bg-slate-900 text-white px-2 py-1 rounded opacity-0 transition-opacity duration-300">Copied</div>'
            f'<textarea id="{textarea_id}" class="absolute left-[-9999px] top-0 opacity-0 pointer-events-none">{escaped_raw}</textarea>'
            f"<pre><code{lang_class}>{code}</code></pre></div>"
        )

    def render_link(self, token):
        href, inner, title = token.target, self.render_inner(token), f' title="{token.title}"' if token.title else ""
        is_hash = href.startswith("#")
        is_external = href.startswith(("http://", "https://", "mailto:", "tel:", "//"))
        is_absolute_internal = href.startswith("/") and not href.startswith("//")
        is_relative = not is_external and not is_absolute_internal
        download_flag = bool(token.title and "download=true" in token.title.lower())
        if "?download=true" in href.lower():
            href = re.sub(r"\?download=true", "", href, flags=re.IGNORECASE)
            download_flag = True
        if is_hash:
            link_class = "text-amber-600 dark:text-amber-400 underline underline-offset-2 hover:text-amber-800 dark:hover:text-amber-200 font-medium transition-colors"
            return f'<a href="{href}" class="{link_class}"{title}>{inner}</a>'
        if is_relative:
            original_href = href
            if href.endswith(".md"):
                href = href[:-3]
            if self.current_path:
                root = get_root_folder().resolve()
                current_dir = (root / self.current_path).parent
                resolved = (current_dir / href).resolve()
                logger.debug(f"DEBUG: original_href={original_href}, current_path={self.current_path}, current_dir={current_dir}, resolved={resolved}, root={root}")
                try:
                    href = f"/posts/{resolved.relative_to(root)}"
                    is_absolute_internal = True
                except ValueError as e:
                    is_external = True
                    logger.debug(f"DEBUG: FAILED - ValueError: {e}")
            else:
                is_external = True
        is_internal = is_absolute_internal and "." not in href.split("/")[-1]
        hx = f' hx-get="{href}" hx-target="#main-content" hx-push-url="true" hx-swap="innerHTML show:window:top"' if is_internal else ""
        ext = "" if (is_internal or is_absolute_internal or is_hash) else ' target="_blank" rel="noopener noreferrer"'
        download_attr = ""
        boost_attr = ""
        if download_flag:
            download_attr = " download"
            boost_attr = ' hx-boost="false"'
            download_target = href[len("/posts/"):] if href.startswith("/posts/") else href.lstrip("/") if href.startswith("/") else href
            href = f"/download/{download_target}"
            hx = ""
        link_class = "text-amber-600 dark:text-amber-400 underline underline-offset-2 hover:text-amber-800 dark:hover:text-amber-200 font-medium transition-colors"
        return f'<a href="{href}"{hx}{ext}{download_attr}{boost_attr} class="{link_class}"{title}>{inner}</a>'


def from_md(content, img_dir=None, current_path=None):
    if img_dir is None and current_path:
        path_parts = Path(current_path).parts
        img_dir = "/posts/" + "/".join(path_parts[:-1]) if len(path_parts) > 1 else "/posts"

    def _protect_escaped_dollar(md):
        code_blocks = []
        def repl(m):
            code_blocks.append(m.group(0))
            return f"__VYASA_CODEBLOCK_{len(code_blocks)-1}__"
        md = re.sub(r"(```+|~~~+)[\s\S]*?\1", repl, md)
        md = re.sub(r"(`+)([^`]*?)\1", repl, md)
        md = re.sub(r"(\\+)\$", lambda m: "\\" * (len(m.group(1)) - 1) + "@@VYASA_DOLLAR@@", md)
        for i, block in enumerate(code_blocks):
            md = md.replace(f"__VYASA_CODEBLOCK_{i}__", block)
        return md

    content = _protect_escaped_dollar(content)
    content, footnotes = extract_footnotes(content)
    content = preprocess_super_sub(content)
    content, code_include_store = preprocess_code_includes(
        content, current_path=current_path, root_folder=get_root_folder()
    )
    content, callout_data_store = preprocess_callouts(content)
    content, tab_data_store = preprocess_md_tabs(content)
    content = preserve_md_newlines(content)
    mods = {
        "pre": "my-4", "p": "text-base leading-relaxed mb-6", "li": "text-base leading-relaxed",
        "ul": "uk-list uk-list-bullet space-y-2 mb-6 ml-6 text-base", "ol": "uk-list uk-list-decimal space-y-2 mb-6 ml-6 text-base",
        "hr": "border-t border-border my-8", "h1": "text-3xl font-bold mb-6 mt-8", "h2": "text-2xl font-semibold mb-4 mt-6",
        "h3": "text-xl font-semibold mb-3 mt-5", "h4": "text-lg font-semibold mb-2 mt-4",
        "table": "uk-table uk-table-striped uk-table-hover uk-table-divider uk-table-middle my-6",
    }
    with ContentRenderer(
        YoutubeEmbed, IframeEmbed, DownloadEmbed, InlineCodeAttr, Strikethrough,
        FootnoteRef, Superscript, Subscript, img_dir=img_dir, footnotes=footnotes,
        current_path=current_path,
    ) as renderer:
        html_out = renderer.render(mst.Document(content))
    if tab_data_store:
        def _render_tab_content(tab_content):
            with ContentRenderer(
                YoutubeEmbed, IframeEmbed, DownloadEmbed, InlineCodeAttr, Strikethrough,
                FootnoteRef, Superscript, Subscript, img_dir=img_dir, footnotes=footnotes,
                current_path=current_path,
            ) as renderer:
                return renderer.render(mst.Document(tab_content))
        html_out = postprocess_md_tabs(html_out, tab_data_store, _render_tab_content)
    if callout_data_store:
        def _render_callout_body(callout_body):
            return _render_markdown_fragment(callout_body, img_dir=img_dir, current_path=current_path)
        for callout_id, callout in callout_data_store.items():
            rendered = _render_callout(callout["kind"], callout["body"], _render_callout_body, title=callout.get("title"), fold=callout.get("fold"))
            placeholder = f'<div class="vyasa-callout-placeholder" data-callout-id="{callout_id}"></div>'
            html_out = html_out.replace(placeholder, rendered)
    if code_include_store:
        for include_id, include in code_include_store.items():
            if include["file_path"].exists():
                text = include["file_path"].read_text(encoding="utf-8")
                line_spec = _parse_line_spec(include["spec"])
                start, end = line_spec if line_spec else (1, len(text.splitlines()))
                lines = text.splitlines()
                snippet = "\n".join(lines[start - 1:end])
                lang = _infer_code_language(include["path_text"])
                hl = _parse_highlight_spec(include["spec"])
                rendered = _render_code_include(snippet, lang=lang, start=start, highlight_spec=hl)
            else:
                rendered = _render_callout(
                    "warning",
                    f'Code include not found: `{include["path_text"]}`',
                    lambda body: mst.markdown(body, partial(ContentRenderer, img_dir=img_dir, current_path=current_path)).strip(),
                )
            placeholder = f'<div class="vyasa-code-include-placeholder" data-include-id="{include_id}"></div>'
            html_out = html_out.replace(placeholder, rendered)
    html_out = re.sub(r"(<table\b[\s\S]*?</table>)", r'<div class="vyasa-table-scroll">\1</div>', html_out, flags=re.IGNORECASE)
    return Div(Link(rel="stylesheet", href=_asset_url("/static/sidenote.css")), NotStr(apply_classes(html_out, class_map_mods=mods)), cls="w-full")



import json

from fasthtml.common import *


def rbac_admin_content(cfg, error, success, preview_text):
    return Div(
        H1("RBAC Administration", cls="text-3xl font-bold"),
        P("Edits save to SQLite immediately and also update the .vyasa file for transparency.", cls="text-slate-600 dark:text-slate-400"),
        P("Rule patterns are matched against request paths (e.g. /posts/ai/...).", cls="text-slate-500 dark:text-slate-500 text-sm"),
        Div(P(error, cls="text-red-600") if error else None, P(success, cls="text-emerald-600") if success else None, cls="mt-4"),
        Form(
            Div(Label(Input(type="checkbox", name="enabled", checked=cfg.get("enabled", False), cls="mr-2"), Span("Enable RBAC"), cls="flex items-center gap-2"), cls="mt-6"),
            Div(Label("Default roles (comma separated)", cls="block text-sm font-medium mb-2"), Input(type="text", name="default_roles", value=", ".join(cfg.get("default_roles", [])), cls="w-full px-3 py-2 rounded-md border border-slate-200 dark:border-slate-700 bg-white/80 dark:bg-slate-900/60"), cls="mt-6"),
            Div(Label("Role users JSON", cls="block text-sm font-medium mb-2"), Textarea(json.dumps(cfg.get("role_users", {}), indent=2, sort_keys=True), name="role_users_json", rows="6", cls="w-full px-3 py-2 font-mono text-xs rounded-md border border-slate-200 dark:border-slate-700 bg-white/80 dark:bg-slate-900/60"), cls="mt-6"),
            Div(Label("User roles JSON", cls="block text-sm font-medium mb-2"), Textarea(json.dumps(cfg.get("user_roles", {}), indent=2, sort_keys=True), name="user_roles_json", rows="6", cls="w-full px-3 py-2 font-mono text-xs rounded-md border border-slate-200 dark:border-slate-700 bg-white/80 dark:bg-slate-900/60"), cls="mt-6"),
            Div(Label("Rules JSON", cls="block text-sm font-medium mb-2"), Textarea(json.dumps(cfg.get("rules", []), indent=2, sort_keys=True), name="rules_json", rows="8", cls="w-full px-3 py-2 font-mono text-xs rounded-md border border-slate-200 dark:border-slate-700 bg-white/80 dark:bg-slate-900/60"), cls="mt-6"),
            Button("Save RBAC", type="submit", cls="mt-6 px-4 py-2 rounded-md bg-blue-600 text-white hover:bg-blue-700"),
            method="post", cls="mt-4",
        ),
        Div(H2("Preview (.vyasa)", cls="text-xl font-semibold mt-10"), Pre(preview_text, cls="mt-3 p-4 rounded-md bg-slate-100 dark:bg-slate-900/60 text-xs overflow-x-auto")),
        cls="max-w-3xl mx-auto py-10 px-6",
    )



from fasthtml.common import RedirectResponse, Response


async def handle_login(request, *, get_config, logger, local_auth_enabled, resolve_roles, rbac_cfg, google_oauth_cfg, coerce_list, login_content, google_oauth_enabled):
    config = get_config()
    user, pwd = config.get_auth()
    logger.info(f"Login attempt for user: {user}")
    error = request.query_params.get("error")
    if request.method == "POST":
        if not local_auth_enabled:
            return RedirectResponse("/login?error=Local+login+disabled", status_code=303)
        form = await request.form()
        username, password = form.get("username", ""), form.get("password", "")
        if username == user and password == pwd:
            roles = resolve_roles({"provider": "local", "username": username}, rbac_cfg, google_oauth_cfg, coerce_list)
            request.session["auth"] = {"provider": "local", "username": username, "roles": roles}
            return RedirectResponse(request.session.pop("next", "/"), status_code=303)
        error = "Invalid username or password."
    return login_content(error, google_oauth_enabled, local_auth_enabled)


async def handle_admin_impersonate(htmx, request, *, get_auth_from_request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list, apply_impersonation_action, resolve_roles, layout, impersonate_content):
    auth = get_auth_from_request(request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list)
    roles = auth.get("roles") if auth else []
    impersonator = auth.get("impersonator") if auth else None
    impersonator_roles = impersonator.get("roles") if impersonator else []
    if (not roles or "full" not in roles) and (not impersonator_roles or "full" not in impersonator_roles):
        return Response("Forbidden", status_code=403)
    error = success = None
    if request.method == "POST":
        form = await request.form()
        error, success = apply_impersonation_action(request.session, request.session.get("auth"), form.get("action", "start"), (form.get("email") or "").strip(), lambda imp_auth: resolve_roles(imp_auth, rbac_cfg, google_oauth_cfg, coerce_list))
    current_auth, impersonator = request.session.get("auth"), request.session.get("impersonator")
    impersonating_email = current_auth.get("email") or current_auth.get("username") if impersonator and current_auth and current_auth.get("provider") == "impersonate" else None
    return layout(impersonate_content(error, success, impersonating_email), htmx=htmx, title="Impersonate", show_sidebar=False, auth=auth, htmx_nav=False)


async def handle_admin_rbac(htmx, request, *, get_auth_from_request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list, parse_rbac_form, parse_roles_text, rbac_db_write, write_rbac_to_vyasa, set_rbac_cfg, cached_build_post_tree, cached_posts_sidebar_html, render_rbac_toml, rbac_admin_content, layout):
    auth = get_auth_from_request(request, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list)
    roles = auth.get("roles") if auth else []
    if not roles or "full" not in roles:
        return Response("Forbidden", status_code=403)
    error = success = None
    cfg = rbac_cfg
    if request.method == "POST":
        form = await request.form()
        new_cfg, error = parse_rbac_form(form, parse_roles_text)
        if not error:
            try:
                rbac_db_write(new_cfg); write_rbac_to_vyasa(new_cfg); set_rbac_cfg(new_cfg); cached_build_post_tree.cache_clear(); cached_posts_sidebar_html.cache_clear(); success = "RBAC settings saved."; cfg = new_cfg
            except Exception as exc:
                error = f"Failed to save RBAC settings: {exc}"
    return layout(rbac_admin_content(cfg, error, success, render_rbac_toml(cfg)), htmx=htmx, title="RBAC Admin", show_sidebar=False, auth=auth, htmx_nav=False)



from pathlib import Path


def asset_url(path: str) -> str:
    """Return static URL with mtime cache-busting token."""
    rel = path.lstrip("/")
    if not rel.startswith("static/"):
        return path
    static_root = Path(__file__).resolve().parent / "static"
    file_path = static_root / rel.replace("static/", "", 1)
    try:
        token = int(file_path.stat().st_mtime_ns)
    except OSError:
        return path
    sep = "&" if "?" in path else "?"
    return f"{path}{sep}v={token}"



import re
from pathlib import Path

from fasthtml.common import *
from monsterui.all import *
from .helpers import resolve_heading_anchor


def collapsible_sidebar(icon, title, items_list, is_open=False, data_sidebar=None, shortcut_key=None, extra_content=None, scroll_target="container"):
    summary_content = [Span(UkIcon(icon, cls="w-5 h-5 block"), cls="flex items-center justify-center w-5 h-5 shrink-0 leading-none"), Span(title, cls="flex-1 leading-none")]
    if shortcut_key:
        summary_content.append(Kbd(shortcut_key, cls="kbd-key px-2.5 py-1.5 text-xs font-mono font-semibold bg-gradient-to-b from-slate-50 to-slate-200 dark:from-slate-700 dark:to-slate-900 text-slate-800 dark:text-slate-200 rounded-md border-2 border-slate-300 dark:border-slate-600 shadow-[0_2px_0_0_rgba(0,0,0,0.1),inset_0_1px_0_0_rgba(255,255,255,0.5)] dark:shadow-[0_2px_0_0_rgba(0,0,0,0.5),inset_0_1px_0_0_rgba(255,255,255,0.1)]"))
    common_frost_style = "bg-white/20 dark:bg-slate-950/70 backdrop-blur-lg border border-slate-900/10 dark:border-slate-700/25 ring-1 ring-white/20 dark:ring-slate-900/30 shadow-[0_24px_60px_-40px_rgba(15,23,42,0.45)] dark:shadow-[0_28px_70px_-45px_rgba(2,6,23,0.85)]"
    summary_classes = f"flex items-center gap-2 font-semibold cursor-pointer py-2.5 px-3 hover:bg-slate-100/80 dark:hover:bg-slate-800/80 rounded-lg select-none list-none {common_frost_style} min-h-[56px]"
    content_classes = f"p-3 {common_frost_style} rounded-lg max-h-[calc(100vh-18rem)] flex flex-col overflow-hidden min-h-0" if scroll_target == "list" else f"p-3 {common_frost_style} rounded-lg overflow-x-auto overflow-y-auto max-h-[calc(100vh-18rem)] sidebar-scroll-container"
    list_classes = "list-none pt-2 sidebar-scroll-container" if scroll_target == "list" else "list-none pt-4"
    extra_content = extra_content or []
    content_id = "sidebar-scroll-container" if scroll_target != "list" else None
    return Details(Summary(*summary_content, cls=summary_classes, style="margin: 0 0 0.5rem 0;"), Div(*extra_content, (Div(Ul(*items_list, cls=list_classes, id=("sidebar-scroll-container" if scroll_target == "list" else None)), cls="min-w-0 flex-1 min-h-0 overflow-x-auto overflow-y-auto") if scroll_target == "list" else Ul(*items_list, cls=list_classes, id=content_id)), cls=content_classes, id=content_id, style="will-change: auto;"), open=is_open, data_sidebar=data_sidebar, style="will-change: auto;")


def extract_toc(content, strip_inline_markdown, text_to_anchor, unique_anchor):
    content_no_code = re.sub(r"^```.*?^```", "", content, flags=re.MULTILINE | re.DOTALL)
    content_no_code = re.sub(r"^~~~.*?^~~~", "", content_no_code, flags=re.MULTILINE | re.DOTALL)
    headings, counts = [], {}
    for match in re.finditer(r"^(#{1,6})\s+(.+)$", content_no_code, flags=re.MULTILINE):
        raw_text = strip_inline_markdown(match.group(2).strip())
        text, anchor = resolve_heading_anchor(raw_text, counts)
        headings.append((len(match.group(1)), text, anchor))
    return headings


def build_toc_items(headings):
    if not headings:
        return [Li("No headings found", cls="text-sm text-slate-500 dark:text-slate-400 py-1")]
    return [Li(A(text, href=f"#{anchor}", cls=f"toc-link block py-1 px-2 text-sm rounded hover:bg-slate-100 dark:hover:bg-slate-800 text-slate-700 dark:text-slate-300 hover:text-blue-600 transition-colors {'ml-0' if level == 1 else f'ml-{(level-1)*3}'}", data_anchor=anchor), cls="my-1") for level, text, anchor in headings]


def get_custom_css_links(root, current_path=None, section_class=None):
    css_elements = []
    for filename in ["custom.css", "style.css"]:
        if (root / filename).exists():
            css_elements.append(Link(rel="stylesheet", href=f"/posts/{filename}"))
            break
    if current_path and section_class:
        post_dir = Path(current_path).parent if "/" in current_path else Path(".")
        for ancestor in ([] if str(post_dir) == "." else [Path(*post_dir.parts[:idx]) for idx in range(1, len(post_dir.parts) + 1)]):
            for filename in ["custom.css", "style.css"]:
                css_file = root / ancestor / filename
                if css_file.exists():
                    css_elements.append(Style(f"\n                        #main-content.{section_class} {{\n                            {css_file.read_text()}\n                        }}\n                    "))
                    break
    return css_elements



from starlette.responses import RedirectResponse, Response

from .policy import is_allowed, normalize_auth, path_requires_roles, resolve_roles


def make_user_auth_before(auth_required, rbac_rules, rbac_cfg, google_oauth_cfg, coerce_list):
    def user_auth_before(req, sess):
        auth = sess.get("auth", None)
        if not auth:
            if auth_required or path_requires_roles(req.url.path, rbac_rules):
                sess["next"] = req.url.path
                return RedirectResponse("/login", status_code=303)
            req.scope["auth"] = None
            return None
        auth = normalize_auth(auth)
        if rbac_rules:
            auth["roles"] = auth.get("roles") or resolve_roles(auth, rbac_cfg, google_oauth_cfg, coerce_list)
            if not is_allowed(req.url.path, auth["roles"], rbac_rules):
                return Response("Forbidden", status_code=403)
        req.scope["auth"] = auth
        return None
    return user_auth_before



import re


def preprocess_tabs(content):
    tab_data_store = {}
    tabs_pattern = re.compile(r"^:::tabs\s*\n(.*?)^:::", re.MULTILINE | re.DOTALL)
    def replace_tabs_block(match):
        tabs = []
        for tab_match in re.finditer(r"^::tab\{([^\}]+)\}\s*\n(.*?)(?=^::tab\{|\Z)", match.group(1), re.MULTILINE | re.DOTALL):
            attrs = {k: v for k, v in re.findall(r'([a-zA-Z0-9_-]+)\s*=\s*"([^"]*)"', tab_match.group(1))}
            if attrs.get("title"):
                tabs.append({"title": attrs["title"], "content": tab_match.group(2).strip(), "attrs": attrs})
        if not tabs:
            return match.group(0)
        title_map = {tab["title"]: tab for tab in tabs}
        index_map = {str(i): tab for i, tab in enumerate(tabs)}
        def resolve_tab_content(tab, stack=None):
            copy_from = tab.get("attrs", {}).get("copy-from")
            if not copy_from:
                return tab["content"]
            stack = stack or set()
            source_tab = index_map.get(copy_from.split(":", 1)[1].strip()) if copy_from.startswith("index:") else index_map.get(copy_from) if copy_from.isdigit() else title_map.get(copy_from)
            if not source_tab or copy_from in stack:
                return tab["content"]
            stack.add(copy_from)
            content = resolve_tab_content(source_tab, stack)
            stack.remove(copy_from)
            fence = "`" * max(4, max((len(run) for run in re.findall(r"`+", content)), default=0) + 1)
            return f"{fence}\n{content}\n{fence}"
        for tab in tabs:
            tab["content"] = resolve_tab_content(tab)
        tab_id = __import__("hashlib").md5(match.group(0).encode()).hexdigest()[:8]
        tab_data_store[tab_id] = [(tab["title"], tab["content"]) for tab in tabs]
        return f'<div class="tab-placeholder" data-tab-id="{tab_id}"></div>'
    return tabs_pattern.sub(replace_tabs_block, content), tab_data_store


def postprocess_tabs(html, tab_data_store, render_tab_content):
    for tab_id, tabs in tab_data_store.items():
        parts = [f'<div class="tabs-container" data-tabs-id="{tab_id}">', '<div class="tabs-header">']
        for i, (title, _) in enumerate(tabs):
            parts.append(f'<button class="tab-button {"active" if i == 0 else ""}" onclick="switchTab(\'{tab_id}\', {i})">{title}</button>')
        parts.append("</div><div class=\"tabs-content\">")
        for i, (_, tab_content) in enumerate(tabs):
            parts.append(f'<div class="tab-panel {"active" if i == 0 else ""}" data-tab-index="{i}">{render_tab_content(tab_content)}</div>')
        parts.append("</div></div>")
        html = html.replace(f'<div class="tab-placeholder" data-tab-id="{tab_id}"></div>', "\n".join(parts))
    return html



from fasthtml.common import *
from monsterui.all import *


def navbar_view(blog_title, theme_toggle_node, show_mobile_menus=False, htmx_nav=True, posts_menu_items=None, compact_mode=False):
    home_link_attrs = {"hx_get": "/", "hx_target": "#main-content", "hx_push_url": "true", "hx_swap": "outerHTML show:window:top settle:0.1s"} if htmx_nav else {}
    left = Div(A(blog_title, href="/", **home_link_attrs), cls="flex items-center gap-2")
    menu = None if not posts_menu_items else Details(Summary(UkIcon("menu", cls="w-4 h-4"), Span("Library"), cls="list-none flex items-center gap-2 cursor-pointer select-none rounded-md px-3 py-2 text-slate-100 hover:bg-slate-800/80 transition-colors [&::-webkit-details-marker]:hidden"), Div(Ul(*posts_menu_items, cls="list-none text-sm max-h-[60vh] overflow-y-auto pr-2"), cls="absolute right-0 mt-2 w-80 p-3 rounded-lg bg-white text-slate-800 shadow-lg border border-slate-200 dark:bg-slate-900 dark:text-slate-100 dark:border-slate-700 z-[1100]"), cls="relative hidden xl:block")
    right = Div(menu, theme_toggle_node, cls="flex items-center gap-3")
    if compact_mode:
        compact_menu = None if not posts_menu_items else Details(Summary(UkIcon("menu", cls="w-5 h-5"), cls="list-none p-2 cursor-pointer rounded hover:bg-slate-800 transition-colors [&::-webkit-details-marker]:hidden"), Div(Ul(*posts_menu_items, cls="list-none text-sm max-h-[60vh] overflow-y-auto pr-2"), cls="absolute left-0 mt-2 w-80 p-3 rounded-lg bg-white text-slate-800 shadow-lg border border-slate-200 dark:bg-slate-900 dark:text-slate-100 dark:border-slate-700 z-[1100]"), cls="relative")
        row = Div(Div(compact_menu or Div(cls="w-9"), cls="w-16"), A(blog_title, href="/", cls="flex-1 px-4 text-center truncate", **home_link_attrs), Div(theme_toggle_node, cls="w-16 flex justify-end"), cls="flex items-center justify-between")
        return Div(row, cls="bg-slate-900 text-white p-4 my-4 rounded-lg shadow-md dark:bg-slate-800")
    if show_mobile_menus:
        mobile = Div(Button(UkIcon("menu", cls="w-5 h-5"), title="Toggle file tree", id="mobile-posts-toggle", cls="p-2 hover:bg-slate-800 rounded transition-colors", type="button"), A(blog_title, href="/", cls="flex-1 px-4 text-center truncate", **home_link_attrs), Div(Button(UkIcon("list", cls="w-5 h-5"), title="Toggle table of contents", id="mobile-toc-toggle", cls="p-2 hover:bg-slate-800 rounded transition-colors", type="button"), theme_toggle_node, cls="flex items-center gap-1"), cls="flex items-center justify-between xl:hidden")
        desktop = Div(left, right, cls="hidden xl:flex items-center justify-between")
        return Div(mobile, desktop, cls="bg-slate-900 text-white p-4 my-4 rounded-lg shadow-md dark:bg-slate-800")
    return Div(left, right, cls="flex items-center justify-between bg-slate-900 text-white p-4 my-4 rounded-lg shadow-md dark:bg-slate-800")



import json


def apply_impersonation_action(session, current_auth, action, email, resolve_roles):
    impersonator = session.get("impersonator")
    if action == "stop":
        if not impersonator:
            return "Not currently impersonating.", None
        session["auth"] = impersonator
        session.pop("impersonator", None)
        return None, "Impersonation stopped."
    if not email:
        return "Email is required.", None
    if not impersonator:
        session["impersonator"] = current_auth
    auth = {"provider": "impersonate", "email": email, "username": email}
    if session.get("impersonator"):
        auth["impersonator"] = session.get("impersonator")
    auth["roles"] = resolve_roles(auth)
    session["auth"] = auth
    return None, f"Now impersonating {email}."


def parse_rbac_form(form, parse_roles_text):
    try:
        role_users = json.loads(form.get("role_users_json", "{}").strip() or "{}")
        user_roles = json.loads(form.get("user_roles_json", "{}").strip() or "{}")
        rules = json.loads(form.get("rules_json", "[]").strip() or "[]")
    except Exception as exc:
        return None, f"Invalid JSON: {exc}"
    if not isinstance(role_users, dict):
        return None, "Role users JSON must be an object."
    if not isinstance(user_roles, dict):
        return None, "User roles JSON must be an object."
    if not isinstance(rules, list):
        return None, "Rules JSON must be an array."
    return {
        "enabled": form.get("enabled") == "on",
        "default_roles": parse_roles_text(form.get("default_roles", "")),
        "role_users": role_users,
        "user_roles": user_roles,
        "rules": rules,
    }, None
